# pylint: disable=redefined-outer-name, unused-argument

import itertools
import operator

import numpy as np
import pytest

import buzzard as buzz
from buzzard.test.tools import assert_tiles_eq
from buzzard.test import make_tile_set

def is_invalid(tup):
    w, h, pox, poy = tup
    return (
        pox == 0 or poy == 0 or w == 0 or h == 0 or
        (pox > w) or
        (poy > h) or
        (w % pox != 0) or
        (h % poy != 0)
    )

PARAMS2 = ['br', 'tr', 'tl', 'bl']
COMBOS = {  # len = 1600
    (w, h, pox, poy)
    for w, h, pox, poy in itertools.product(range(4), range(4), range(10), range(10))
}
FAIL_COMBOS = {  # len = 1575
    tup for tup in COMBOS if is_invalid(tup)
}
VALID_COMBOS = COMBOS - FAIL_COMBOS  # len = 25

RANDOM_SIZES = [796, 795]
RANDOM_OCCURRENCES = [2, 3, 4]
RANDOM_SRC_SIZE = [2000, 794]
RANDOM_COMBOS = {
    tup
    for tup in itertools.product(
        RANDOM_SRC_SIZE, RANDOM_SRC_SIZE,
        RANDOM_SIZES, RANDOM_SIZES,
        RANDOM_OCCURRENCES, RANDOM_OCCURRENCES,
        PARAMS2,
    )
}
RANDOM_FAIL_COMBOS = {
    tup for tup in RANDOM_COMBOS if is_invalid(tup[2:6])
}
RANDOM_VALID_COMBOS = RANDOM_COMBOS - RANDOM_FAIL_COMBOS

# *************************************************************************** **
# FIXTURES ****************************************************************** **
# *************************************************************************** **

@pytest.fixture(scope='module')
def fps():
    """
    See make_tile_set
    A B C D E F G
    H I J K L M N
    O P Q.R.S T U
    V W X.Y.Z a b
    c d e.f.g h i
    j k l m n o p
    q r s t u v w
    """
    return make_tile_set.make_tile_set(7, [1, -1], [1, -1])


def pytest_generate_tests(metafunc):
    if metafunc.function == test_fail:
        metafunc.parametrize(
            argnames='w, h, pox, poy',
            argvalues=FAIL_COMBOS
        )
    if metafunc.function == test_success:
        metafunc.parametrize(
            argnames='w, h, pox, poy',
            argvalues=VALID_COMBOS,
        )
    if metafunc.function == test_random_fail:
        metafunc.parametrize(
            argnames='srcw, srch, w, h, pox, poy, boundary_effect_locus',
            argvalues=RANDOM_FAIL_COMBOS,
        )
    if metafunc.function == test_random_success:
        metafunc.parametrize(
            argnames='srcw, srch, w, h, pox, poy, boundary_effect_locus',
            argvalues=RANDOM_VALID_COMBOS,
        )

# *************************************************************************** **
# TESTS  ******************************************************************** **
# *************************************************************************** **

def test_random_success(srcw, srch, w, h, pox, poy, boundary_effect_locus):
    fp = buzz.Footprint(tl=(0, 0), rsize=(srcw, srch), size=(srcw, srch))
    tiles = fp.tile_occurrence((w, h), pox, poy,
                               boundary_effect_locus=boundary_effect_locus)
    for f in SUCCESS_ASSERTS:
        f(fp, tiles, (w, h), pox, poy, boundary_effect_locus)

def test_random_fail(srcw, srch, w, h, pox, poy, boundary_effect_locus):
    fp = buzz.Footprint(tl=(0, 0), rsize=(srcw, srch), size=(srcw, srch))
    with pytest.raises(ValueError):
        fp.tile_occurrence((w, h), pox, poy, boundary_effect_locus=boundary_effect_locus)


def test_fail(fps, w, h, pox, poy):
    with pytest.raises(ValueError):
        fps.Qg.tile_occurrence((w, h), pox, poy)


def test_success(fps, w, h, pox, poy):
    # Tiles of area 1
    if (1, 1, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.Q, fps.R, fps.S],
            [fps.X, fps.Y, fps.Z],
            [fps.e, fps.f, fps.g],
        ]

    # Tiles of area 2
    elif (2, 1, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.QR, fps.ST],
            [fps.XY, fps.Za],
            [fps.ef, fps.gh],
        ]
    elif (2, 1, 2, 1) == (w, h, pox, poy):
        truth = [
            [fps.PQ, fps.QR, fps.RS, fps.ST],
            [fps.WX, fps.XY, fps.YZ, fps.Za],
            [fps.de, fps.ef, fps.fg, fps.gh],
        ]
    elif (1, 2, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.QX, fps.RY, fps.SZ],
            [fps.el, fps.fm, fps.gn],
        ]
    elif (1, 2, 1, 2) == (w, h, pox, poy):
        truth = [
            [fps.JQ, fps.KR, fps.LS],
            [fps.QX, fps.RY, fps.SZ],
            [fps.Xe, fps.Yf, fps.Zg],
            [fps.el, fps.fm, fps.gn],
        ]

    # Tiles of area 3
    elif (1, 3, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.Qe, fps.Rf, fps.Sg, ],
        ]
    elif (1, 3, 1, 3) == (w, h, pox, poy):
        truth = [
            [fps.CQ, fps.DR, fps.ES, ],
            [fps.JX, fps.KY, fps.LZ, ],
            [fps.Qe, fps.Rf, fps.Sg, ],
            [fps.Xl, fps.Ym, fps.Zn, ],
            [fps.es, fps.ft, fps.gu, ],
        ]
    elif (3, 1, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.QS, ],
            [fps.XZ, ],
            [fps.eg, ],
        ]
    elif (3, 1, 3, 1) == (w, h, pox, poy):
        truth = [
            [fps.OQ, fps.PR, fps.QS, fps.RT, fps.SU, ],
            [fps.VX, fps.WY, fps.XZ, fps.Ya, fps.Zb, ],
            [fps.ce, fps.df, fps.eg, fps.fh, fps.gi, ],
        ]

    # Tiles of area 4
    elif (2, 2, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.QY, fps.Sa],
            [fps.em, fps.go],
        ]
    elif (2, 2, 2, 1) == (w, h, pox, poy):
        truth = [
            [fps.PX, fps.QY, fps.RZ, fps.Sa],
            [fps.dl, fps.em, fps.fn, fps.go],
        ]
    elif (2, 2, 1, 2) == (w, h, pox, poy):
        truth = [
            [fps.JR, fps.LT],
            [fps.QY, fps.Sa],
            [fps.Xf, fps.Zh],
            [fps.em, fps.go],
        ]
    elif (2, 2, 2, 2) == (w, h, pox, poy):
        truth = [
            [fps.IQ, fps.JR, fps.KS, fps.LT],
            [fps.PX, fps.QY, fps.RZ, fps.Sa],
            [fps.We, fps.Xf, fps.Yg, fps.Zh],
            [fps.dl, fps.em, fps.fn, fps.go],
        ]

    # Tiles of area 6
    elif (2, 3, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.Qf, fps.Sh],
        ]
    elif (2, 3, 2, 1) == (w, h, pox, poy):
        truth = [
            [fps.Pe, fps.Qf, fps.Rg, fps.Sh, ],
        ]
    elif (2, 3, 1, 3) == (w, h, pox, poy):
        truth = [
            [fps.CR, fps.ET, ],
            [fps.JY, fps.La, ],
            [fps.Qf, fps.Sh, ],
            [fps.Xm, fps.Zo, ],
            [fps.et, fps.gv, ],
        ]
    elif (2, 3, 2, 3) == (w, h, pox, poy):
        truth = [
            [fps.BQ, fps.CR, fps.DS, fps.ET, ],
            [fps.IX, fps.JY, fps.KZ, fps.La, ],
            [fps.Pe, fps.Qf, fps.Rg, fps.Sh, ],
            [fps.Wl, fps.Xm, fps.Yn, fps.Zo, ],
            [fps.ds, fps.et, fps.fu, fps.gv, ],
        ]
    elif (3, 2, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.QZ],
            [fps.en],
        ]
    elif (3, 2, 1, 2) == (w, h, pox, poy):
        truth = [
            [fps.JS, ],
            [fps.QZ, ],
            [fps.Xg, ],
            [fps.en, ],
        ]
    elif (3, 2, 3, 1) == (w, h, pox, poy):
        truth = [
            [fps.OX, fps.PY, fps.QZ, fps.Ra, fps.Sb, ],
            [fps.cl, fps.dm, fps.en, fps.fo, fps.gp, ],
        ]
    elif (3, 2, 3, 2) == (w, h, pox, poy):
        truth = [
            [fps.HQ, fps.IR, fps.JS, fps.KT, fps.LU, ],
            [fps.OX, fps.PY, fps.QZ, fps.Ra, fps.Sb, ],
            [fps.Ve, fps.Wf, fps.Xg, fps.Yh, fps.Zi, ],
            [fps.cl, fps.dm, fps.en, fps.fo, fps.gp, ],
        ]

    # Tiles of area 9
    elif (3, 3, 1, 1) == (w, h, pox, poy):
        truth = [
            [fps.Qg],
        ]
    elif (3, 3, 1, 3) == (w, h, pox, poy):
        truth = [
            [fps.CS, ],
            [fps.JZ, ],
            [fps.Qg, ],
            [fps.Xn, ],
            [fps.eu, ],
        ]
    elif (3, 3, 3, 1) == (w, h, pox, poy):
        truth = [
            [fps.Oe, fps.Pf, fps.Qg, fps.Rh, fps.Si, ],
        ]
    elif (3, 3, 3, 3) == (w, h, pox, poy):
        truth = [
            [fps.AQ, fps.BR, fps.CS, fps.DT, fps.EU],
            [fps.HX, fps.IY, fps.JZ, fps.Ka, fps.Lb],
            [fps.Oe, fps.Pf, fps.Qg, fps.Rh, fps.Si],
            [fps.Vl, fps.Wm, fps.Xn, fps.Yo, fps.Zp],
            [fps.cs, fps.dt, fps.eu, fps.fv, fps.gw],
        ]

    else:
        raise Exception('Test %s not implemented' % str((w, h, pox, poy)))
    tiles = fps.Qg.tile_occurrence((w, h), pox, poy)
    assert_tiles_eq(tiles, truth)
    for f in SUCCESS_ASSERTS:
        f(fps.Qg, tiles, (w, h), pox, poy, 'tl')


def assert_property_tile_size(src, tiles, size, occx, occy, boundary_effect_locus):
    w = np.vectorize(operator.attrgetter('w'))(tiles.flatten())
    assert np.unique(w).size == 1
    h = np.vectorize(operator.attrgetter('h'))(tiles.flatten())
    assert np.unique(h).size == 1

def assert_property_pixel_coverage(src, tiles, size, occx, occy, boundary_effect_locus):
    mask = np.zeros(src.shape, dtype='int')
    tiles = tiles.flatten()
    for t in tiles:
        mask[t.slice_in(src, clip=True)] += 1
    assert (mask == occx * occy).all()

def assert_property_share_area(src, tiles, size, occx, occy, boundary_effect_locus):
    border_tiles = np.r_[tiles[-1, 1:-1], tiles[0, 1:-1], tiles[:, 0], tiles[:, -1]]
    for t in border_tiles:
        assert t.share_area(src)

def assert_property_unique(src, tiles, size, occx, occy, boundary_effect_locus):
    tls = np.vectorize(operator.attrgetter('tl'), signature='()->(2)')(tiles.flatten())
    assert np.unique(tls, axis=0).shape[0] == tiles.size


SUCCESS_ASSERTS = [
    assert_property_tile_size,
    assert_property_pixel_coverage,

    assert_property_share_area,
    assert_property_unique,
]

def test_value_error(fps):
    with pytest.raises(ValueError, match='shape'):
        fps.AI.tile_occurrence(1, 1, 1)
    with pytest.raises(ValueError, match='shape'):
        fps.AI.tile_occurrence([1, 1, 1], 1, 1)
    with pytest.raises(ValueError, match='effect'):
        fps.AI.tile_occurrence((1, 1), 1, 1, boundary_effect='')
    with pytest.raises(ValueError, match='effect_locus'):
        fps.AI.tile_occurrence((1, 1), 1, 1, boundary_effect_locus='')



# pylint: disable=redefined-outer-name
import numpy as np
import pytest
import buzzard as buzz

S = 2 ** 14

def strigify(fp):
    return '{:<25} {!s:<45} {:<25} tl:{!s:<35} tr:{!s:<35} br:{!s:<35} bl:{!s:<35} {} {}'.format(
        fp.angle,
        fp.scale.tolist(),
        np.divide.reduce(fp.scale),
        fp.tl,
        fp.tr,
        fp.br,
        fp.bl,
        str(fp.aff6).replace('\n', ''),
        fp.rsize,
    )

def pytest_generate_tests(metafunc):
    with buzz.Env(allow_complex_footprint=1, significant=10):
        fp0 = buzz.Footprint(rsize=(S, S), size=(S, S), tl=(50000, 50000))
        transfos = [
            (fp0,
             fp0,
             'identity',
            ),
            (fp0,
             buzz.Footprint(rsize=(S, S), size=(S * 2, S * 2), tl=(50000, 50000)),
             'double unit',
            ),
            (fp0,
             buzz.Footprint(rsize=(S, S), gt=(50000, 1, 0, 50000, 0, -2)),
             'double y unit',
            ),
            (fp0,
             fp0.intersection(fp0, rotation=45).clip(0, 0, S, S),
             'rotation 45',
            ),
            (fp0,
             buzz.Footprint(rsize=(S, S), size=(S * 2, S * 2), tl=(50000, 50000)).intersection(fp0.dilate(S), rotation=45).clip(0, 0, S, S),
             'rotation 45 and double unit'
            ),
            (fp0,
             buzz.Footprint(rsize=(S, S), gt=(50000, 1, 0, 50000, 0, -2)).intersection(fp0.dilate(S), rotation=45).clip(0, 0, S, S),
             'rotation 45 and double y unit'
            ),
            (fp0,
             buzz.Footprint(rsize=(S, S), gt=(50000, 1, 0, 50000, 0, 1)),
             'mirror',
            ),
        ]

    tests = []
    for src_fp, trg_fp, transfo_name in transfos:
        assert np.all(src_fp.rsize == trg_fp.rsize), (src_fp.rsize, trg_fp.rsize)

        for scalex in [1, -1]:
            for scaley in [1, -1]:
                for rot in [0, 45, -45, 90, -90, -135, 135, 180, -180]:
                    for noise_factor in [0, 1e-6]:
                        tests.append((
                            transfo_name, src_fp, trg_fp, scalex, scaley, rot, noise_factor
                        ))
    metafunc.parametrize(
        argnames='transfo_name,src_fp,trg_fp,scalex,scaley,rot,noise_factor',
        argvalues=tests,
    )

def test_move(transfo_name, src_fp, trg_fp, scalex, scaley, rot, noise_factor):
    with buzz.Env(allow_complex_footprint=1, significant=10):

        src_fp = src_fp.dilate(S).intersection(
            src_fp.dilate(S), rotation=src_fp.angle + rot, scale=src_fp.scale * [scalex, scaley]
        ).clip(0, 0, S, S)
        trg_fp = trg_fp.dilate(S).intersection(
            trg_fp.dilate(S), rotation=trg_fp.angle + rot, scale=trg_fp.scale * [scalex, scaley]
        ).clip(0, 0, S, S)

        print('// src_properties:', strigify(src_fp))
        print('// trg_properties:', strigify(trg_fp))
        assert np.all(src_fp.rsize == trg_fp.rsize)

        print('->{} noise at {:.1e}'.format(transfo_name, noise_factor))
        tl, tr, br = trg_fp.tl, trg_fp.tr, trg_fp.br
        tl, tr, br = np.asarray([tl, tr, br]) + np.random.rand(3, 2) * noise_factor

        out_fp0 = src_fp.move(tl, tr, br)
        print('  without snap:   ', strigify(out_fp0))

        # Assert that `Footprint.move` does no give strange results
        assert trg_fp.almost_equals(out_fp0)

        out_fp1 = src_fp.move(tl, tr, br, True)
        print('  with snap:      ', strigify(out_fp1))

        # Assert that `new function` does no give strange results
        assert trg_fp.almost_equals(out_fp1)

        if src_fp.angle == trg_fp.angle and np.isclose(0, (src_fp.angle + 360) % 90):
            # Assert that angles are be fully preserved when rotation if a multiple of 90
            assert out_fp1.angle == src_fp.angle



import enum
import collections
import itertools
import logging
import os

from buzzard._actors.message import Msg
from buzzard._actors.cached.query_infos import CacheComputationInfos

LOGGER = logging.getLogger(__name__)

class ActorCacheSupervisor(object):
    """Actor that takes care of tracking, checking and scheduling computation of cache files"""

    def __init__(self, raster):
        """
        Parameter
        ---------
        raster: _a_recipe_raster.ABackRecipeRaster
        """
        self._raster = raster
        self._cache_fps_status = {
            cache_fp: _CacheTileStatus.unknown
            for cache_fp in raster.cache_fps.flat
        }
        self._queries = {}
        self._alive = True
        self.address = '/Raster{}/CacheSupervisor'.format(self._raster.uid)
        self._directory_primed = False

        # Should contain the path to all files that will be opened using the Dataset's activation
        # pool. It means all cache files in those status:
        # - _CacheTileStatus.checking
        # - _CacheTileStatus.ready
        self._path_of_cache_fp = raster.async_dict_path_of_cache_fp

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_make_those_cache_files_available(self, qi):
        """Receive message: Ensure that the cache files for this query can be read, if necessary
        create the missing ones, but launch at most one collection process. If several are missing,
        compute those in the same order as the query needs it.

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """

        if not self._directory_primed:
            self._directory_primed = True
            os.makedirs(self._raster.cache_dir, exist_ok=True)
            if self._raster.overwrite:
                file_list = self._raster.list_cache_path_candidates()
                LOGGER.info('Removing {} cache files'.format(
                    len(file_list)
                ))
                for path in file_list:
                    os.remove(path)

        msgs = []
        cache_fps = qi.list_of_cache_fp

        query = _Query()
        self._queries[qi] = query

        for cache_fp in cache_fps:
            status = self._cache_fps_status[cache_fp]

            if status == _CacheTileStatus.ready:
                query.cache_fps_ensured.add(cache_fp)

            elif status == _CacheTileStatus.checking:
                query.cache_fps_checking.add(cache_fp)

            elif status == _CacheTileStatus.absent:
                query.cache_fps_to_compute.add(cache_fp)

            elif status == _CacheTileStatus.unknown:
                path_candidates = self._raster.list_cache_path_candidates(cache_fp)
                if len(path_candidates) == 1:
                    self._cache_fps_status[cache_fp] = _CacheTileStatus.checking
                    self._path_of_cache_fp[cache_fp] = path_candidates[0]
                    self._raster.debug_mngr.event('cache_file_update', self._raster.facade_proxy, cache_fp, 'unknown')
                    query.cache_fps_checking.add(cache_fp)
                    msgs += [
                        Msg('FileChecker', 'infer_cache_file_status', cache_fp, path_candidates[0])
                    ]
                else:
                    self._cache_fps_status[cache_fp] = _CacheTileStatus.absent
                    self._raster.debug_mngr.event('cache_file_update', self._raster.facade_proxy, cache_fp, 'absent')
                    for path in path_candidates: # pragma: no cover
                        LOGGER.warning(
                            'Removing {} because {} tiles with the same prefix'.format(path, len(path_candidates))
                        )
                        os.remove(path)
                    query.cache_fps_to_compute.add(cache_fp)
            else: # pragma: no cover
                assert False

        if len(query.cache_fps_ensured) != 0:
            # Notify the production pipeline that those cache tiles are already ready
            msgs += [
                Msg('CacheExtractor', 'cache_files_ready', {
                    fp: self._path_of_cache_fp[fp]
                    for fp in query.cache_fps_ensured
                })
            ]
        if len(query.cache_fps_checking) == 0:
            # CacheSupervisor is now done working on this query
            del self._queries[qi]

            if len(query.cache_fps_to_compute) > 0:
                # Some tiles need to be computed and none need to be checked, launching collection right
                # now
                msgs += self._query_start_collection(qi, query)

        return msgs

    def receive_inferred_cache_file_status(self, cache_fp, path, status):
        """Receive message: One cache tile was checked

        Parameters
        ----------
        cache_fp: Footprint
        path: str
        status: bool
        """
        msgs = []

        # assertions
        assert self._cache_fps_status[cache_fp] == _CacheTileStatus.checking
        for query in self._queries.values():
            assert cache_fp not in query.cache_fps_ensured
            assert cache_fp not in query.cache_fps_to_compute

        if status:
            # This cache tile is OK to be read
            # - notify the production pipeline
            self._path_of_cache_fp[cache_fp] = path
            self._cache_fps_status[cache_fp] = _CacheTileStatus.ready
            self._raster.debug_mngr.event('cache_file_update', self._raster.facade_proxy, cache_fp, 'ready')
            msgs += [
                Msg('CacheExtractor', 'cache_files_ready', {cache_fp: path})
            ]
        else:
            # This cache tile was corrupted and removed
            self._cache_fps_status[cache_fp] = _CacheTileStatus.absent
            del self._path_of_cache_fp[cache_fp]
            self._raster.debug_mngr.event('cache_file_update', self._raster.facade_proxy, cache_fp, 'absent')

        queries_treated = []
        for qi, query in self._queries.items():
            if cache_fp in query.cache_fps_checking:
                query.cache_fps_checking.remove(cache_fp)
                if status:
                    query.cache_fps_ensured.add(cache_fp)
                else:
                    query.cache_fps_to_compute.add(cache_fp)

                if len(query.cache_fps_checking) == 0:
                    # CacheSupervisor is now done working on this query
                    queries_treated.append(qi)

                    if len(query.cache_fps_to_compute) > 0:
                        # Some tiles need to be computed and none need to be checked, launching collection right
                        # now
                        msgs += self._query_start_collection(qi, query)

        for qi in queries_treated:
            del self._queries[qi]

        return msgs

    def receive_cache_file_written(self, cache_fp, path):
        """Receive message: One cache file was just written to disk

        Parameters
        ----------
        cache_fp: Footprint
        path: str
        """
        msgs = []
        assert self._cache_fps_status[cache_fp] == _CacheTileStatus.absent

        self._path_of_cache_fp[cache_fp] = path
        self._cache_fps_status[cache_fp] = _CacheTileStatus.ready
        self._raster.debug_mngr.event('cache_file_update', self._raster.facade_proxy, cache_fp, 'ready')
        msgs += [
            Msg('CacheExtractor', 'cache_files_ready', {cache_fp: path})
        ]
        return msgs

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        if qi in self._queries:
            del self._queries[qi]
        return []

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        self._queries.clear()
        self._path_of_cache_fp = None
        self._cache_fps_status.clear()
        self._raster = None
        return []

    # ******************************************************************************************* **
    def _query_start_collection(self, qi, query):
        assert len(query.cache_fps_checking) == 0
        assert len(query.cache_fps_to_compute) > 0
        cache_fps = [
            fp
            for fp in qi.list_of_cache_fp
            if fp in query.cache_fps_to_compute
        ]
        assert qi.cache_computation is None
        qi.cache_computation = CacheComputationInfos(qi, self._raster, cache_fps)
        self._raster.debug_mngr.event('object_allocated', qi.cache_computation)
        return [
            Msg('/Global/GlobalPrioritiesWatcher', 'a_query_need_those_cache_tiles',
                self._raster.uid, qi, cache_fps
            ),
            Msg('ComputationGate1', 'compute_those_cache_files', qi),
        ]

    # ******************************************************************************************* **

class _CacheTileStatus(enum.Enum):
    unknown = 0
    checking = 1
    absent = 2
    ready = 3

class _Query(object):
    def __init__(self):
        self.cache_fps_checking = set()
        self.cache_fps_ensured = set()
        self.cache_fps_to_compute = set()



import numpy as np

from buzzard._actors.message import Msg

class ActorComputationGate2(object):
    """Actor that takes care of delaying the computation of a cache file until inputs are ready."""

    def __init__(self, raster):
        self._raster = raster
        self._queries = {}
        self._alive = True
        self.address = '/Raster{}/ComputationGate2'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_compute_this_array(self, qi, compute_idx):
        """Receive message: Wait for the inputs of this computation to be ready

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        compute_idx: int
        """
        msgs = []

        qicc = qi.cache_computation
        assert qicc is not None

        if qi in self._queries:
            q = self._queries[qi]
            assert q.allowed_up_count == compute_idx, ''
        else:
            assert compute_idx == 0, 'This is the first call for qi, compute_idx it should be 0'
            q = _Query()
            self._queries[qi] = q
        q.allowed_up_count = compute_idx + 1

        msgs += self._allow(qi, q)
        return msgs

    def receive_input_queue_update(self, queue_key):
        """Receive message: One of the input queues of a query changed in size.

        Parameters
        ----------
        queue_key: (_actors.cached.query_infos.QueryInfos, str)
        """
        msgs = []

        qi, _ = queue_key
        if qi in self._queries:
            q = self._queries[qi]
            msgs += self._allow(qi, q)

        return msgs

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        if qi in self._queries:
            del self._queries[qi]
        return []

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False
        self._queries.clear()
        self._raster = None
        return []

    # ******************************************************************************************* **
    @staticmethod
    def _allow(qi, q):
        msgs = []
        qicc = qi.cache_computation

        if len(qicc.primitive_queue_per_primitive) > 0:
            qsizes = [v.qsize() for v in qicc.primitive_queue_per_primitive.values()]
            min_qsize = min(qsizes)
            allowed_gate2_count = qicc.collected_count + min_qsize
            assert q.allowed_down_count <= allowed_gate2_count, (
                'allowed more than ready count'
            )
        else:
            allowed_gate2_count = np.inf

        i = q.allowed_down_count
        while i < allowed_gate2_count and i < q.allowed_up_count:
            msgs += [Msg(
                'Computer', 'compute_this_array', qi, i,
            )]
            i += 1
        q.allowed_down_count = i

        return msgs

    # ******************************************************************************************* **

class _Query(object):
    def __init__(self):
        self.allowed_up_count = 0 # How many compute allowed by `ComputationGate1`
        self.allowed_down_count = 0 # How many allowed to `Computer`



import uuid
import queue
import weakref

from buzzard._a_source_raster import ASourceRaster, ABackSourceRaster
from buzzard._footprint import Footprint
from buzzard import _tools
from buzzard._actors.message import Msg
from buzzard._debug_observers_manager import DebugObserversManager

QUEUE_POLL_DISTANCE = 0.1

class AAsyncRaster(ASourceRaster):
    """Base abstract class defining the common behavior of all rasters that are managed by the
    Dataset's scheduler.

    Features Defined
    ----------------
    - Has a `queue_data`, a low level method that can be used to query several arrays at once.
    - Has an `iter_data`, a higher level wrapper of `queue_data`.
    """

    def queue_data(self, fps, channels=None, dst_nodata=None, interpolation='cv_area',
                   max_queue_size=5, **kwargs):
        """Read several rectangles of data on several channels from the source raster.

        Using `queue_data` instead of multiple calls to `get_data` allows more parallelism.
        The `fps` parameter should contain a sequence of `Footprint` that will be mapped to
        `numpy.ndarray`. The first ones will be computed with a higher priority than the later ones.

        Calling this method sends an asynchronous message to the Dataset's scheduler with the
        input parameters and a queue. On the input side of the queue, the scheduler will call the
        `put` method with each array requested. On the output side of the queue, the `get` method
        should be called to retrieve the requested arrays.

        The output queue will be created with a max queue size of `max_queue_size`, the scheduler
        will be careful to prepare only the arrays that can fit in the output queue. Thanks to this
        feature: backpressure can be entirely avoided.

        If you wish to cancel your request, loose the reference to the queue and the scheduler will
        gracefuly cancel the query.

        In general you should use the `iter_data` method instead of the `queue_data` one, it is much
        safer to use. However you will need to pass the `queue_data` method of a raster, to create
        another raster (a recipe) that depends on the first raster.

        see  rasters' `get_data` documentation, it shares most of the concepts

        Parameters
        ----------
        fps: sequence of Footprint
            The Footprints at which the raster should be sampled.
        channels:
            see `get_data` method
        dst_nodata:
            see `get_data` method
        interpolation:
            see `get_data` method
        max_queue_size: int
            Maximum number of arrays to prepare in advance in the underlying queue.

        Returns
        -------
        queue: queue.Queue of ndarray
            The arrays are put into the queue in the same order as in the `fps` parameter.

        """
        for fp in fps:
            if not isinstance(fp, Footprint):
                msg = 'element of `fps` parameter should be a Footprint (not {})'.format(fp) # pragma: no cover
                raise ValueError(msg)

        return self._back.queue_data(
            fps=fps,
            parent_uid=None,
            key_in_parent=None,
            **_tools.parse_queue_data_parameters(
                'queue_data', self, channels, dst_nodata, interpolation, max_queue_size, **kwargs
            )
        )

    def iter_data(self, fps, channels=None, dst_nodata=None, interpolation='cv_area',
                  max_queue_size=5, **kwargs):
        """Read several rectangles of data on several channels from the source raster.

        The `iter_data` method is a higher level wrapper around the `queue_data` method. It
        returns a python generator and while waiting for data, it periodically probes the
        Dataset's scheduler to reraise an exception if it crashed.

        If you wish to cancel your request, loose the reference to the iterable and the scheduler
        will gracefully cancel the query.

        see rasters' `get_data` documentation, it shares most of the concepts
        see `queue_data` documentation, it is called from within the `iter_data` method

        Parameters
        ----------
        fps: sequence of Footprint
            The Footprints at which the raster should be sampled.
        channels:
            see `get_data` method
        dst_nodata:
            see `get_data` method
        interpolation:
            see `get_data` method
        max_queue_size: int
            Maximum number of arrays to prepare in advance in the underlying queue.

        Returns
        -------
        iterable: iterable of ndarray
            The arrays are yielded into the generator in the same order as in the `fps` parameter.

        """
        for fp in fps:
            if not isinstance(fp, Footprint):
                raise ValueError('element of `fps` parameter should be a Footprint (not {})'.format(
                    fp
                )) # pragma: no cover

        return self._back.iter_data(
            fps=fps,
            **_tools.parse_queue_data_parameters(
                'iter_data', self, channels, dst_nodata, interpolation, max_queue_size, **kwargs
            )
        )

class ABackAsyncRaster(ABackSourceRaster):
    """Implementation of AAsyncRaster's specifications"""

    def __init__(self, resample_pool, max_resampling_size, debug_observers, **kwargs):
        self.uid = uuid.uuid4()
        self.resample_pool = resample_pool
        self.max_resampling_size = max_resampling_size
        self.debug_mngr = DebugObserversManager(debug_observers)

        # Quick hack to share the dict of path to cache files with the ActorCacheSupervisor
        # This is currently needed to perform the `.close` operation
        # This is a clear violation of the separation of concerns
        self.async_dict_path_of_cache_fp = {}

        super().__init__(**kwargs)

    def queue_data(self, fps, channel_ids, dst_nodata, interpolation, max_queue_size, is_flat,
                   parent_uid, key_in_parent):
        q = queue.Queue(max_queue_size)
        self.back_ds.put_message(Msg(
            '/Raster{}/QueriesHandler'.format(self.uid),
            'new_query',
            weakref.ref(q),
            max_queue_size,
            fps,
            channel_ids,
            is_flat,
            dst_nodata,
            interpolation,
            parent_uid,
            key_in_parent
        ))
        return q

    def iter_data(self, fps, channel_ids, dst_nodata, interpolation, max_queue_size, is_flat):
        q = self.queue_data(fps, channel_ids, dst_nodata, interpolation, max_queue_size, is_flat,
                            None, None)
        def _iter_data_generator():
            i = 0
            while True:
                try:
                    while i < len(fps):
                        arr = q.get(True, timeout=QUEUE_POLL_DISTANCE)
                        yield arr
                        i += 1
                    return
                except queue.Empty:
                    timeout = True
                else:
                    timeout = False
                if timeout:
                    self.back_ds.ensure_scheduler_still_alive()
        return _iter_data_generator()

    def get_data(self, fp, channel_ids, dst_nodata, interpolation):
        it = self.iter_data(
            [fp], channel_ids, dst_nodata, interpolation, 1,
            False, # `is_flat` is not important since caller reshapes output
        )
        return next(it)

    def create_actors(self): # pragma: no cover
        raise NotImplementedError('ABackAsyncRaster.create_actors is virtual pure')

    def close(self):
        """Virtual method:
        - May be overriden
        - Should always be called

        Should be called after scheduler's end
        """
        self.back_ds.put_message(Msg(
            '/Global/TopLevel', 'kill_raster', self,
        ), check_scheduler_status=False)
        # TODO: just sending a kill_raster message may not be enough. Need synchro?
        self.back_ds.deactivate_many(self.async_dict_path_of_cache_fp.values())
        super().close()



""" Stochastic test that proves that MultiOrderedDict behaves like _MultiOrderedDict_NSquared
Since _MultiOrderedDict_NSquared implementation is straightforward we can trust it, and
transfer this confidence to MultiOrderedDict using this test.
"""

import weakref
import gc
import collections
import uuid
import itertools

import pytest
import numpy as np

from buzzard._tools import MultiOrderedDict

class _MultiOrderedDict_NSquared(object):
    """Class with the same specifications as MultiOrderedDict but with a simpler and less effective
    implementation. It exsits for unit testing purposes"""

    def __init__(self):
        self._l = list()

    def __len__(self):
        return len(self._l)

    def __contains__(self, key):
        for k, _ in self._l:
            if k == key:
                return True
        else:
            return False

    def count(self, key):
        return sum(
            key == k
            for k, _ in self._l
        )

    def pop_back(self):
        if len(self._l) == 0:
            collections.OrderedDict().popitem(last=True)
        return self._l.pop(-1)

    def push_front(self, key, value):
        self._l.insert(0, (key, value))

    def pop_first_occurrence(self, key):
        for i, (k, v) in enumerate(self._l):
            if k == key:
                break
        else:
            raise KeyError('{} not in MultiOrderedDict'.format(key))
        self._l.pop(i)
        return v

    def pop_last_occurrence(self, key):
        for i, (k, v) in list(enumerate(self._l))[::-1]:
            if k == key:
                break
        else:
            raise KeyError('{} not in MultiOrderedDict'.format(key))
        self._l.pop(i)
        return v

    def pop_all_occurrences(self, key):
        indices = []
        res = []
        for i, (k, v) in enumerate(self._l):
            if k == key:
                indices.append(i)
        for i in sorted(indices, reverse=True):
            res.append(self._l.pop(i)[1])
        return res


class Value():
    def __init__(self, i, h):
        self.i = i
        self.h = h

    def __eq__(self, other):
        return self.i == other.i

    def __hash__(self):
        return self.h.int

    def __str__(self):
        return '<{}, {}>'.format(
            self.i, self.h
        )

    def __repr__(self):
        return str(self)


def _assert(ref, ref_res, test_res):
    if isinstance(ref_res, Exception):
        assert isinstance(test_res, Exception), (ref_res, test_res, ref._l)
        assert type(ref_res) is type(test_res) and ref_res.args == test_res.args, (ref_res, test_res, ref._l)
    else:
        assert ref_res == test_res, (ref_res, test_res, ref._l)

def _assert_collection(ref_wset, test_wset):
    if len(ref_wset) != len(test_wset):
        gc.collect()
    assert ref_wset == test_wset, (set(ref_wset), set(test_wset))

def _test_a(ref, test, ref_wset, test_wset, rng, verbose=False):
    k = rng.randint(0, 5)
    v = rng.randint(9999)
    h = uuid.uuid4()

    vref = Value(v, h)
    vtest = Value(v, h)
    assert vref == vtest

    if verbose:
        print('{:>20}: k:{:5} v:{:5}'.format("push_front", k, v))

    ref.push_front(k, vref)
    test.push_front(k, vtest)

    ref_wset.add(vref)
    test_wset.add(vtest)
    _assert_collection(ref_wset, test_wset)

def _test_b(ref, test, _ref_wset, _test_wset, _rng, verbose=False):
    try:
        ref_res = ref.pop_back()
    except Exception as e:
        ref_res = e
    try:
        test_res = test.pop_back()
    except Exception as e:
        test_res = e

    if verbose:
        if isinstance(test_res, Exception):
            print('{:>20}:                     {!r}'.format("pop_back", test_res))
        else:
            k, v = test_res
            print('{:>20}: k:{:5} v:{!s:5}'.format("pop_back", k, v))
    _assert(ref, ref_res, test_res)

def _test_c(ref, test, ref_wset, test_wset, rng, verbose=False):
    k = rng.randint(0, 5)
    try:
        ref_res = ref.pop_first_occurrence(k)
    except Exception as e:
        ref_res = e
    try:
        test_res = test.pop_first_occurrence(k)
    except Exception as e:
        test_res = e

    if verbose:
        if isinstance(test_res, Exception):
            print('{:>20}:                     {!r}'.format("pop_first_occurrence", test_res))
        else:
            print('{:>20}: k:{:5} v:{!s:5}'.format("pop_first_occurrence", k, test_res))
    _assert(ref, ref_res, test_res)
    _assert_collection(ref_wset, test_wset)

def _test_d(ref, test, ref_wset, test_wset, rng, verbose=False):
    k = rng.randint(0, 5)
    try:
        ref_res = ref.pop_last_occurrence(k)
    except Exception as e:
        ref_res = e
    try:
        test_res = test.pop_last_occurrence(k)
    except Exception as e:
        test_res = e

    if verbose:
        if isinstance(test_res, Exception):
            print('{:>20}:                     {!r}'.format("pop_last_occurrence", test_res))
        else:
            print('{:>20}: k:{:5} v:{!s:5}'.format("pop_last_occurrence", k, test_res))
    _assert(ref, ref_res, test_res)
    _assert_collection(ref_wset, test_wset)

def _test_e(ref, test, _ref_wset, _test_wset, rng, _verbose=False):
    k = rng.randint(0, 5)
    assert (k in ref) == (k in test)

def _test_f(ref, test, _ref_wset, _test_wset, _rng, _verbose=False):
    assert len(ref) == len(test)

def _test_g(ref, test, ref_wset, test_wset, rng, verbose=False):
    k = rng.randint(0, 5)
    try:
        ref_res = ref.pop_all_occurrences(k)
    except Exception as e:
        ref_res = e
    try:
        test_res = test.pop_all_occurrences(k)
    except Exception as e:
        test_res = e

    if verbose:
        if isinstance(test_res, Exception):
            print('{:>20}:                     {!r}'.format("pop_all_occurrences", test_res))
        else:
            print('{:>20}: k:{:5} v:{!s:5}'.format("pop_all_occurrences", k, test_res))
    _assert(ref, ref_res, test_res)
    _assert_collection(ref_wset, test_wset)

def _test_h(ref, test, _ref_wset, _test_wset, rng, _verbose=False):
    k = rng.randint(0, 5)
    assert ref.count(k) == test.count(k)


def test_multi_ordered_dict():
    ref = _MultiOrderedDict_NSquared()
    test = MultiOrderedDict()
    ref_wset = weakref.WeakSet()
    test_wset = weakref.WeakSet()
    rng = np.random.RandomState()

    tests = [
        _test_a,
        _test_b,
        _test_c,
        _test_d,
        _test_e,
        _test_f,
        _test_g,
        _test_h,
    ]

    for _ in range(5000):
        i = rng.randint(0, len(tests))
        tests[i](ref, test, ref_wset, test_wset, rng, False)



import collections

class DebugObserversManager(object):
    """Delivers the callbacks to the observers provided by user in the `debug_observers` parameters.
    """
    def __init__(self, debug_observers):
        self._obs = debug_observers
        self._to_call_per_ename = _ToCallPerEventName(debug_observers)

    def event(self, ename, *args):
        for method in self._to_call_per_ename[ename]:
            method(*args)

class _ToCallPerEventName(dict):
    def __init__(self, debug_observers):
        self._obs = debug_observers

    def __missing__(self, ename):
        method_name = 'on_{}'.format(ename)
        return [
            getattr(o, method_name)
            for o in self._obs
            if hasattr(o, method_name)
        ]



# pylint: disable=redefined-outer-name, unused-argument

import itertools

import pytest
import buzzard as buzz
import numpy as np

try:
    import torch
except ImportError:
    pytest.skip('No pytorch', allow_module_level=True)

@pytest.fixture(scope="module", autouse=True)
def env():
    np.set_printoptions(linewidth=400, threshold=np.iinfo('int64').max, suppress=True)
    with buzz.Env(allow_complex_footprint=1):
        yield

def pytest_generate_tests(metafunc):
    return _pytest_generate_tests(metafunc)

@buzz.Env(allow_complex_footprint=1)
def _pytest_generate_tests(metafunc):

    # *********************************************************************** **
    scales = [
        (1, 1),
        (1, -1),
        (2, 3),
    ]
    rots = [
        0,
        90, 180, -135,
    ]
    if 'output_padding' in metafunc.fixturenames:
        rsizes = [
            (4, 5),
            (6, 7),
        ]
    else:
        rsizes = [
            (15, 16),
            (17, 18),
        ]
    src_fp = buzz.Footprint(rsize=(1, 1), gt=(100, 1, 0, 100, 0, 1))
    confs = list(itertools.product(scales, rots, rsizes))
    fp0s = []

    for (scalex, scaley), rot, rsize in confs:
        fp = src_fp.dilate(rsize[0] * 3).intersection(
            src_fp.dilate(rsize[0] * 3),
            rotation=rot,
            scale=src_fp.scale * [scalex, scaley]
        ).clip(0, 0, *rsize)
        assert np.allclose(fp.scale, (scalex, scaley))
        assert np.allclose((fp.angle + 360) % 360, (rot + 360) % 360)
        assert np.all(fp.rsize == rsize)
        fp0s.append(fp)

    # *********************************************************************** **
    kernel_sizes = np.asarray([
        (1, 2),
        (3, 4),
    ])

    # *********************************************************************** **
    strides = np.asarray([
        (1, 2),
        (3, 4),
    ])

    # *********************************************************************** **
    if 'output_padding' in metafunc.fixturenames:
        paddings = np.asarray([
            (0, 1),
            (2, 3),
        ])
    else:
        paddings = np.asarray([
            (0, 1),
            (2, 3),
        ])

    # *********************************************************************** **
    dilations = np.asarray([
        (1, 1),
        (2, 3),
    ])

    # *********************************************************************** **
    output_paddings = np.asarray([
        (0, 1),
        (2, 3),
    ])

    # *********************************************************************** **
    if 'output_padding' in metafunc.fixturenames:
        tests = list(itertools.product(
            fp0s, kernel_sizes, strides, paddings, dilations, output_paddings,
        ))
        tests = [
            (fp0, kernel_size, stride, padding, dilation, output_padding)
            for fp0, kernel_size, stride, padding, dilation, output_padding in tests
            if np.all(output_padding < kernel_size) and np.all(output_padding < dilation)
        ]
        metafunc.parametrize(
            argnames='fp0,kernel_size,stride,padding,dilation,output_padding',
            argvalues=tests,
        )
    else:
        metafunc.parametrize(
            argnames='fp0,kernel_size,stride,padding,dilation',
            argvalues=list(itertools.product(
                fp0s, kernel_sizes, strides, paddings, dilations,
            )),
        )

def test_conv2d(fp0, kernel_size, stride, padding, dilation):
    # Test forward against meshgrid computed using pytorch ****************** **
    fp1 = fp0.forward_conv2d(kernel_size, stride, padding, dilation)
    truth = meshgrid_of_forward_conv2d(fp0, kernel_size, stride, padding, dilation)
    assert np.allclose(truth, fp1.meshgrid_spatial)

    # Test backward against forward ***************************************** **
    # (modulo small bottom-right overflows caused by stride and fp1.rsize)
    fp0_bis = fp1.backward_conv2d(kernel_size, stride, padding, dilation)

    angle = (fp0.angle + 360) % 90
    if np.allclose(np.cos(angle * 4), 1):
        # is orthogonal
        assert np.all(fp0.gt == fp0_bis.gt)
    else:
        assert np.allclose(fp0.gt, fp0_bis.gt)

    shape_overflow = fp0.shape - fp0_bis.shape
    assert np.all(shape_overflow >= 0)
    assert np.all(shape_overflow < stride)

def test_convtranspose2d(fp0, kernel_size, stride, padding, dilation, output_padding):
    # Test forward against meshgrid computed with pytorch ******************* **
    # (modulo the meshgrid values at output_padding)
    fp1 = fp0.forward_convtranspose2d(kernel_size, stride, padding, dilation, output_padding)
    truth = meshgrid_of_forward_convtranspose2d(
        fp0, kernel_size, stride, padding, dilation, output_padding,
    )
    fromfp = np.asarray(fp1.meshgrid_spatial)
    fromfp = fromfp[:, :fp1.shape[0] - output_padding[0], :fp1.shape[1] - output_padding[1]]
    assert np.allclose(truth, fromfp)

    # Test backward against forward ***************************************** **
    fp0_bis = fp1.backward_convtranspose2d(kernel_size, stride, padding, dilation, output_padding)

    angle = (fp0.angle + 360) % 90
    if np.allclose(np.cos(angle * 4), 1):
        # is orthogonal
        assert fp0 == fp0_bis
    else:
        assert fp0.almost_equals(fp0_bis)

# *********************************************************************************************** **
# Utils ***************************************************************************************** **
def meshgrid_of_forward_conv2d(fp, kernel_size, stride, padding, dilation):
    fp_padded = fp.dilate(*np.flipud(padding))

    c = torch.nn.Conv2d(
        in_channels=2, out_channels=2, groups=2, bias=False, padding=0,
        kernel_size=kernel_size, stride=stride.tolist(), dilation=dilation.tolist(),
    )
    c.weight = torch.nn.Parameter(
        torch.ones_like(c.weight) / np.prod(kernel_size),
        requires_grad=False,
    )

    x = np.asarray(fp_padded.meshgrid_spatial).reshape(1, 2, *fp_padded.shape).astype('float32')
    with torch.no_grad():
        y = c(torch.from_numpy(x)).numpy()
    return y.reshape(2, *y.shape[2:])

def create_bilinear_kernel(stride, dtype=None):
    # Parameters ************************************************************ **
    v = stride
    v = np.asarray(v).reshape(-1)
    if v.size == 1:
        v = np.asarray((v, v))
    if v.size != 2:
        raise ValueError('{} should have size 1 or 2'.format(k))
    w = v.astype(int, copy=False)
    if np.any(v != w):
        raise ValueError('{} should be of type int'.format(k))
    if np.any(v < 1):
        raise ValueError('{} should be greater or equal to 1'.format(k))
    stride = v

    # Build kernel ********************************************************** **
    shape = (stride - 1) * 2 + 1
    center = stride - (shape % 2 + 1) / 2
    kernel = np.zeros(shape, dtype='float32')
    for yx in np.ndindex(*shape):
        val = 1 - np.abs((yx - center) / stride)
        kernel[yx[0], yx[1]] = np.prod(val)

    # Assert correctness **************************************************** **
    indices = np.asarray(list(np.ndindex(*kernel.shape)))
    for i in range(stride[0]):
        for j in range(stride[1]):
            mask = (indices % stride == (i, j)).all(axis=1)
            facts = kernel.reshape(-1)[mask]
            assert np.allclose(facts.sum(), 1)

    return kernel

def inner_meshgrid_of_forward_convtranspose2d(fp, stride):
    kernel = create_bilinear_kernel(stride)
    kernel_shape = np.asarray(kernel.shape)
    assert np.all(kernel_shape % 2 == 1)
    padding_input = np.asarray((kernel_shape - 1) // 2)
    padding_parameter = (kernel_shape - 1) - padding_input

    # Assert that the `padding_parameter` is good *************************** **
    c = torch.nn.ConvTranspose2d(
        in_channels=1, out_channels=1, groups=1, bias=False, padding=padding_parameter.tolist(),
        kernel_size=kernel.shape, stride=stride.tolist(),
    )
    c.weight = torch.nn.Parameter(
        torch.ones_like(c.weight) * torch.from_numpy(kernel), requires_grad=False
    )
    x = np.ones(shape=(1, 1, *fp.shape), dtype='float32')
    with torch.no_grad():
        y = c(torch.from_numpy(x)).numpy()
    assert np.allclose(y, 1)
    del x, y

    # Compute the meshgrid ************************************************** **
    c = torch.nn.ConvTranspose2d(
        in_channels=2, out_channels=2, groups=2, bias=False, padding=padding_parameter.tolist(),
        kernel_size=kernel.shape, stride=stride.tolist(),
    )
    c.weight = torch.nn.Parameter(
        torch.from_numpy(np.stack([kernel, kernel])[:, None, :, :]), requires_grad=False
    )

    x = np.asarray(fp.meshgrid_spatial).reshape(1, 2, *fp.shape).astype('float32')
    with torch.no_grad():
        y = c(torch.from_numpy(x)).numpy()
    return y[0]

def meshgrid_of_forward_convtranspose2d(fp, kernel_size, stride, padding, dilation, output_padding):
    dilation_parameter = dilation
    kernel_size_parameter = kernel_size
    padding_parameter = padding
    del padding, dilation

    kernel_size = 1 + (kernel_size_parameter - 1) * dilation_parameter
    padding_input = dilation_parameter * (kernel_size_parameter - 1) - padding_parameter
    todilate_input = np.ceil(padding_input / stride).astype(int)
    tocrop_inner_meshgrid = todilate_input * stride - padding_input
    fp_padded = fp.dilate(*np.flipud(todilate_input))

    # Create inner meshgrid ************************************************* **
    # First create a larger one and then crop what's needed
    mg = inner_meshgrid_of_forward_convtranspose2d(fp_padded, stride)
    mg = mg[
        :,
        tocrop_inner_meshgrid[0]:mg.shape[1] - tocrop_inner_meshgrid[0],
        tocrop_inner_meshgrid[1]:mg.shape[2] - tocrop_inner_meshgrid[1],
    ]

    # Apply conv2d on inner meshgrid **************************************** **
    c = torch.nn.Conv2d(
        in_channels=2, out_channels=2, groups=2, bias=False, padding=0, stride=1, dilation=1,
        kernel_size=kernel_size,
    )
    c.weight = torch.nn.Parameter(
        torch.ones_like(c.weight) / np.prod(kernel_size),
        requires_grad=False,
    )

    x = mg[None, :, :, :]
    with torch.no_grad():
        y = c(torch.from_numpy(x)).numpy()
    mg = y[0]

    # Check meshgrid shape ************************************************** **
    truth_shape = output_shape_of_conv2dtranspose(
        fp,
        kernel_size=kernel_size_parameter, stride=stride,
        padding=padding_parameter, dilation=dilation_parameter,
        output_padding=output_padding,
    )
    found_shape = mg.shape[1:]
    found_shape = tuple(found_shape + output_padding)
    assert found_shape == truth_shape

    return mg

def output_shape_of_conv2dtranspose(fp, **kwargs):
    c = torch.nn.ConvTranspose2d(
        in_channels=1, out_channels=1, groups=1, bias=False,
        **{
            k: v.tolist()
            for k, v in kwargs.items()
        },
    )
    x = np.asarray(fp.meshgrid_spatial[0]).reshape(1, 1, *fp.shape).astype('float32')
    with torch.no_grad():
        y = c(torch.from_numpy(x)).numpy()
    return y.shape[2:]



import collections
import sys

import shapely.geometry as sg
import numpy as np

from buzzard._a_source import ASource, ABackSource
from buzzard import _tools
from buzzard._footprint import Footprint
from buzzard._tools import conv

class ASourceVector(ASource):
    """Base abstract class defining the common behavior of all vectors.

    Features Defined
    ----------------
    - Has a `type` that defines the type of geometry (like "Polygon")
    - Has `fields` that define the type of informations that is paired with each geometries
    - Has a `stored` extent that allows to retrieve the current extent of all the geometries
    - Has a length that indicates how many geometries this source contains.
    - Has several read functions (like `iter_data`) to retrieve geometries in their current state to \
        shapely objects
    """

    @property
    def type(self):
        """Geometry type"""
        return self._back.type

    @property
    def fields(self):
        """Fields definition"""
        return [dict(d) for d in self._back.fields]

    @property
    def extent(self):
        """Get the vector's extent in work spatial reference. (`x` then `y`)

        Example
        -------
        >>> minx, maxx, miny, maxy = ds.roofs.extent
        """
        return self._back.extent

    @property
    def extent_stored(self):
        """Get the vector's extent in stored spatial reference. (minx, miny, maxx, maxy)"""
        return self._back.extent_stored

    @property
    def bounds(self):
        """Get the vector's bounds in work spatial reference. (`min` then `max`)

        Example
        -------
        >>> minx, miny, maxx, maxy = ds.roofs.extent
        """
        return self._back.bounds

    @property
    def bounds_stored(self):
        """Get the vector's bounds in stored spatial reference. (`min` then `max`)"""
        return self._back.bounds_stored

    def __len__(self):
        """Return the number of features in vector"""
        return len(self._back)

    def iter_data(self, fields=None, geom_type='shapely',
                  mask=None, clip=False, slicing=slice(0, None, 1)):
        """.. _vector file iter_data:

        Create an iterator over vector's features

        Parameters
        ----------
        fields: None or string or -1 or sequence of string/int
            Which fields to include in iteration

            - if None, empty sequence or empty string: No fields included
            - if -1: All fields included
            - if string: Name of fields to include (separated by comma or space)
            - if sequence: List of indices / names to include

        geom_type: {'shapely', 'coordinates'}
            Returned geometry type
        mask: None or Footprint or shapely geometry or (nbr, nbr, nbr, nbr)
            Add a spatial filter to iteration, only geometries not disjoint with mask will be \
            included.

            - if None: No spatial filter
            - if Footprint or shapely polygon: Polygon
            - if (nbr, nbr, nbr, nbr): Extent (minx, maxx, miny, maxy)

        clip: bool
            Returns intersection of geometries and mask.
            Caveat: A clipped geometry might not be of the same type as the original geometry.
            e.g: polygon might be clipped to might be converted to one of those:

            - polygon
            - line
            - point
            - multipolygon
            - multiline
            - multipoint
            - geometrycollection

        slicing: slice
            Slice of the iteration to return. It is applied after spatial filtering

        Yields
        ------
        feature: geometry or (geometry,) or (geometry, *fields)
            - If `geom_type` is 'shapely': geometry is a `shapely geometry`.
            - If `geom_type` is `coordinates`: geometry is a `nested lists of numpy arrays`.
            - If `fields` is not a sequence: `feature` is `geometry` or `(geometry, *fields)`, \
                 depending on the number of fields to yield.
            - If `fields` is a sequence or a string: `feature` is `(geometry,)` or \
                `(geometry, *fields)`. Use `fields=[-1]` to get a monad containing all fields.

        Examples
        --------
        >>> for polygon, volume, stock_type in ds.stocks.iter_data('volume,type'):
                print('area:{}m**2, volume:{}m**3'.format(polygon.area, volume))

        >>> for polygon, in ds.stocks.iter_data([]):
                print('area:{}m**2'.format(polygon.area))

        >>> for polygon in ds.stocks.iter_data():
                print('area:{}m**2'.format(polygon.area))

        """
        # Normalize and check fields parameter
        field_indices, is_flat = _tools.normalize_fields_parameter(
            fields, self._back.index_of_field_name
        )
        del fields

        # Normalize and check geom_type parameter
        if geom_type not in ['shapely', 'coordinates']: # pragma: no cover
            raise ValueError('Bad parameter `geom_type`')

        # Normalize and check clip parameter
        clip = bool(clip)
        if mask is None and clip is True: # pragma: no cover
            raise ValueError('`clip` is True but `mask` is None')

        # Normalize and check mask parameter
        # TODO: Convert to_work
        mask_poly, mask_rect = self._normalize_mask_parameter(mask)
        del mask

        # Normalize and check slicing parameter
        if not isinstance(slicing, slice): # pragma: no cover
            raise TypeError('`slicing` of type `{}` instead of `slice'.format(
                type(slicing),
            ))

        for data in self._back.iter_data(geom_type, field_indices, slicing,
                                         mask_poly, mask_rect, clip):
            if is_flat:
                assert len(data) == 1, len(data)
                yield data[0]
            else:
                yield data

    def get_data(self, index, fields=-1, geom_type='shapely', mask=None, clip=False):
        """Fetch a single feature in vector. See ASourceVector.iter_data"""
        index = int(index)
        for val in self.iter_data(fields, geom_type, mask, clip, slice(index, index + 1, 1)):
            return val
        else: # pragma: no cover
            raise IndexError('Feature `{}` not found'.format(index))

    def iter_geojson(self, mask=None, clip=False, slicing=slice(0, None, 1)):
        """Create an iterator over vector's features

        Parameters
        ----------
        mask: None or Footprint or shapely geometry or (nbr, nbr, nbr, nbr)
            Add a spatial filter to iteration, only geometries not disjoint with mask will be
            included.

            - if None: No spatial filter
            - if Footprint or shapely polygon: Polygon
            - if (nbr, nbr, nbr, nbr): Extent (minx, maxx, miny, maxy)

        clip: bool
            Returns intersection of geometries and mask.
            Caveat: A clipped geometry might not be of the same type as the original geometry.
            e.g: polygon might be clipped to might be converted to one of those:

            - polygon
            - line
            - point
            - multipolygon
            - multiline
            - multipoint
            - geometrycollection

        slicing: slice
            Slice of the iteration to return. It is applied after spatial filtering

        Returns
        -------
        iterable of geojson feature (dict)


        Example
        -------
        >>> for geojson in ds.stocks.iter_geojson():
                print('exterior-point-count:{}, volume:{}m**3'.format(
                    len(geojson['geometry']['coordinates'][0]),
                    geojson['properties']['volume']
                ))
        """
        # Normalize and check clip parameter
        clip = bool(clip)
        if mask is None and clip is True: # pragma: no cover
            raise ValueError('`clip` is True but `mask` is None')

        # Normalize and check mask parameter
        mask_poly, mask_rect = self._normalize_mask_parameter(mask)
        del mask

        # Normalize and check slicing parameter
        if not isinstance(slicing, slice): # pragma: no cover
            raise TypeError('`slicing` of type `{}` instead of `slice'.format(
                type(slicing),
            ))

        gen = self._back.iter_data(
            'geojson',
            list(range(len(self.fields))),
            slicing,
            mask_poly,
            mask_rect,
            clip,
        )
        for data in gen:
            yield {
                'type': 'Feature',
                'properties': collections.OrderedDict(
                    (field['name'], value)
                    for field, value in zip(self.fields, data[1:])
                ),
                'geometry':  data[0],
            }

    def get_geojson(self, index, mask=None, clip=False):
        """Fetch a single feature in vector. See ASourceVector.iter_geojson"""
        index = int(index)
        for val in self.iter_geojson(mask, clip, slice(index, index + 1, 1)):
            return val
        else: # pragma: no cover
            raise IndexError('Feature `{}` not found'.format(index))

    @staticmethod
    def _normalize_mask_parameter(mask):
        if isinstance(mask, sg.base.BaseGeometry):
            return mask, None
        elif isinstance(mask, Footprint):
            return mask.poly, None
        elif isinstance(mask, collections.Container):
            mask = [float(v) for v in mask]
            minx, maxx, miny, maxy = mask
            mask = minx, miny, maxx, maxy
            return None, mask
        elif mask is None:
            return None, None
        else: # pragma: no cover
            raise TypeError('`mask` should be a Footprint, an extent or a shapely object')

    # Deprecation
    extent_origin = _tools.deprecation_pool.wrap_property(
        'extent_stored',
        '0.4.4'
    )

class ABackSourceVector(ABackSource):
    """Implementation of ASourceVector's specifications"""

    def __init__(self, type, fields, **kwargs):
        super(ABackSourceVector, self).__init__(**kwargs)
        self.type = type
        self.fields = fields
        self.index_of_field_name = {
            field['name']: i
            for i, field in enumerate(self.fields)
        }
        self.type_of_field_index = [
            conv.type_of_oftstr(field['type'])
            for field in self.fields
        ]
        self.all_nullable = all(field['nullable'] for field in self.fields)

    @property
    def extent(self): # pragma: no cover
        raise NotImplementedError('ABackSourceVector.extent is virtual pure')

    @property # pragma: no cover
    def extent_stored(self):
        raise NotImplementedError('ABackSourceVector.extent_stored is virtual pure')

    @property
    def bounds(self):
        extent = self.extent
        return np.asarray([extent[0], extent[2], extent[1], extent[3]])

    @property
    def bounds_stored(self):
        extent = self.extent_stored
        return np.asarray([extent[0], extent[2], extent[1], extent[3]])

    def __len__(self): # pragma: no cover
        raise NotImplementedError('ABackSourceVector.__len__ is virtual pure')

    def iter_data(self, geom_type, field_indices, slicing, mask_poly, mask_rect, clip): # pragma: no cover
        raise NotImplementedError('ABackSourceVector.iter_data is virtual pure')

if sys.version_info < (3, 6):
    # https://www.python.org/dev/peps/pep-0487/
    for k, v in ASourceVector.__dict__.items():
        if hasattr(v, '__set_name__'):
            v.__set_name__(ASourceVector, k)



import functools
import collections
import multiprocessing as mp
import multiprocessing.pool

import numpy as np

from buzzard._actors.message import Msg
from buzzard._actors.pool_job import ProductionJobWaiting, PoolJobWorking
from buzzard._a_source_raster_remap import ABackSourceRasterRemapMixin

class ActorResampler(object):
    """Actor that takes care of resampling sample tiles, and wait for all
    resamplings to be performed for a production array.
    """

    def __init__(self, raster):
        self._raster = raster
        self._alive = True
        resample_pool = raster.resample_pool
        if resample_pool is not None:
            self._waiting_room_address = '/Pool{}/WaitingRoom'.format(id(resample_pool))
            self._working_room_address = '/Pool{}/WorkingRoom'.format(id(resample_pool))
            if isinstance(resample_pool, mp.pool.ThreadPool):
                self._same_address_space = True
            elif isinstance(resample_pool, mp.pool.Pool):
                self._same_address_space = False
            else: # pragma: no cover
                assert False, 'Type should be checked in facade'
        self._waiting_jobs = set()
        self._working_jobs = set()

        self._prod_infos = (
            collections.defaultdict(dict)
        ) # type: Mapping[CachedQueryInfos, Mapping[int, _ProdArray]]

        self.address = '/Raster{}/Resampler'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_resample_and_accumulate(self, qi, prod_idx, sample_fp, resample_fp, subsample_array):
        """Receive message: A resampling operation is ready to be performed

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        prod_idx: int
        sample_fp: None or Footprint of shape (Y, X)
        resample_fp: Footprint of shape (Y', X')
        subsample_array: None or ndarray of shape (Y, X)
        """
        msgs = []

        pi = qi.prod[prod_idx]

        if prod_idx not in self._prod_infos[qi]:
            self._prod_infos[qi][prod_idx] = _ProdArray(pi)
        pr = self._prod_infos[qi][prod_idx]

        is_tile_alone = len(pi.resample_fps) == 1
        is_tile_outside = sample_fp is None
        is_interpolation_needed = not is_tile_outside and not pi.same_grid
        is_interpolation_defered = self._raster.resample_pool is not None

        if is_tile_alone:
            assert pr.arr is None

        if is_tile_alone and is_tile_outside:
            # Case 1: production footprint is fully outside of raster
            assert sample_fp is None
            assert subsample_array is None
            pr.arr = np.full(
                np.r_[resample_fp.shape, len(qi.channel_ids)],
                qi.dst_nodata, self._raster.dtype,
            )
            pr.commit(resample_fp)
            pr.is_post_processed = True

        elif is_tile_alone and sample_fp.almost_equals(pi.fp):
            # Case 2: production footprint is aligned and fully inside raster
            assert subsample_array.shape[:2] == tuple(resample_fp.shape)
            pr.arr = subsample_array
            if self._raster.nodata is not None and self._raster.nodata != qi.dst_nodata:
                pr.arr[pr.arr == self._raster.nodata] = qi.dst_nodata
            pr.arr = pr.arr.astype(self._raster.dtype, copy=False)
            pr.arr = _reorder_channels(qi, pr.arr)

            pr.commit(resample_fp)
            pr.is_post_processed = True

        elif is_tile_alone and not is_interpolation_needed:
            # Case 3: production footprint is aligned and is both inside and outside raster
            pr.arr = np.full(
                np.r_[resample_fp.shape, len(qi.unique_channel_ids)],
                qi.dst_nodata, self._raster.dtype,
            )
            slices = sample_fp.slice_in(resample_fp)
            pr.arr[slices] = subsample_array
            if self._raster.nodata is not None and self._raster.nodata != qi.dst_nodata:
                pr.arr[slices][pr.arr[slices] == self._raster.nodata] = qi.dst_nodata
            pr.arr = _reorder_channels(qi, pr.arr)
            pr.commit(resample_fp)
            pr.is_post_processed = True

        elif not is_tile_alone and is_tile_outside:
            # Case 4: production footprint is fully outside
            pr.commit(resample_fp)

        elif not is_tile_alone and not is_interpolation_needed:
            assert False, 'Without interpolation, there should be only one tile'

        elif is_interpolation_needed and not is_interpolation_defered:
            # Case 5: production footprint not aligned, interpolation on scheduler
            job = self._create_interpolation_work_job(
                qi, prod_idx, sample_fp, resample_fp, subsample_array,
            )
            job.func()
            self._commit_interpolation_work_result(job, None)

        elif is_interpolation_needed and is_interpolation_defered:
            # Case 6: production footprint not aligned, interpolation on pool
            wait = Wait(self, qi, prod_idx, sample_fp, resample_fp, subsample_array)
            self._waiting_jobs.add(wait)
            msgs += [
                Msg(self._waiting_room_address, 'schedule_job', wait)
            ]

        else:
            assert False, """
            is_tile_alone = {}
            is_tile_outside = {}
            is_interpolation_needed = {}
            is_interpolation_defered = {}
            """.format(
                is_tile_alone, is_tile_outside, is_interpolation_needed, is_interpolation_defered
            )

        msgs += self._push_if_done(qi, prod_idx)

        return msgs

    def receive_token_to_working_room(self, job, token):
        """Receive message: Waiting job can proceed to the working room"""
        self._waiting_jobs.remove(job)

        work = self._create_interpolation_work_job(
            job.qi, job.prod_idx, job.sample_fp, job.resample_fp, job.subsample_array,
        )
        self._working_jobs.add(work)

        return [
            Msg(self._working_room_address, 'launch_job_with_token', work, token)
        ]

    def receive_job_done(self, job, result):
        self._working_jobs.remove(job)
        self._commit_interpolation_work_result(job, result)
        return self._push_if_done(job.qi, job.prod_idx)

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        msgs = []

        # Cancel waiting jobs
        jobs_to_kill = [
            job
            for job in self._waiting_jobs
            if job.qi == qi
        ]
        for job in jobs_to_kill:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
            self._waiting_jobs.remove(job)

        # Cancel working jobs
        jobs_to_kill = [
            job
            for job in self._working_jobs
            if job.qi == qi
        ]
        for job in jobs_to_kill:
            msgs += [Msg(self._working_room_address, 'cancel_job', job)]
            self._working_jobs.remove(job)

        if qi in self._prod_infos:
            del self._prod_infos[qi]

        return msgs

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        msgs = []
        for job in self._waiting_jobs:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
        for job in self._working_jobs:
            msgs += [Msg(self._working_room_address, 'cancel_job', job)]
        self._waiting_jobs.clear()
        self._working_jobs.clear()
        self._prod_infos.clear()
        self._raster = None
        return msgs

    # ******************************************************************************************* **
    def _create_interpolation_work_job(self, qi, prod_idx, sample_fp, resample_fp, subsample_array):
        pi = qi.prod[prod_idx]
        pr = self._prod_infos[qi][prod_idx]

        if pr.arr is None:
            pr.arr = np.full(
                np.r_[pi.fp.shape, len(qi.channel_ids)],
                qi.dst_nodata, self._raster.dtype,
            )

        return Work(
            self, qi, prod_idx,
            sample_fp, resample_fp,
            subsample_array, pr.arr
        )

    def _commit_interpolation_work_result(self, work_job, res):
        qi = work_job.qi
        prod_idx = work_job.prod_idx
        resample_fp = work_job.resample_fp
        pr = self._prod_infos[qi][prod_idx]

        pr.commit(resample_fp)

        if self._raster.resample_pool is not None and not self._same_address_space:
            work_job.dst_array_slice[:] = res
        else:
            assert res is None

    def _push_if_done(self, qi, prod_idx):
        msgs = []

        pr = self._prod_infos[qi][prod_idx]
        if pr.done:
            assert pr.arr is not None
            if not pr.is_post_processed:
                pr.arr = _reorder_channels(qi, pr.arr)
                pr.is_post_processed
            msgs += [
                Msg('Producer', 'made_this_array', qi, prod_idx, pr.arr)
            ]
            del self._prod_infos[qi][prod_idx]
            if len(self._prod_infos[qi]) == 0:
                del self._prod_infos[qi]

        return msgs

    # ******************************************************************************************* **

class _ProdArray(object):
    def __init__(self, pi):
        self.arr = None
        self.is_post_processed = False
        self._missing_resample_fps = set(pi.resample_fps)

    def commit(self, resample_fp):
        self._missing_resample_fps.remove(resample_fp)

    @property
    def done(self):
        return len(self._missing_resample_fps) == 0

def _reorder_channels(qi, arr):
    if qi.channel_ids != qi.unique_channel_ids:
        indices = [
            qi.unique_channel_ids.index(bi)
            for bi in qi.channel_ids
        ]
        arr = arr[..., indices]
    return arr

class Wait(ProductionJobWaiting):

    def __init__(self, actor, qi, prod_idx, sample_fp, resample_fp, subsample_array):
        self.qi = qi
        self.prod_idx = prod_idx
        self.sample_fp = sample_fp
        self.resample_fp = resample_fp
        self.subsample_array = subsample_array
        super().__init__(actor.address, qi, prod_idx, 0, self.resample_fp)

class Work(PoolJobWorking):
    def __init__(self, actor, qi, prod_idx, sample_fp, resample_fp, subsample_array, dst_array):
        self.qi = qi
        self.prod_idx = prod_idx
        self.resample_fp = resample_fp
        produce_fp = qi.prod[prod_idx].fp

        dst_array_slice = dst_array[resample_fp.slice_in(produce_fp)]

        if actor._raster.resample_pool is None or actor._same_address_space:
            func = functools.partial(
                _resample_subsample_array,
                sample_fp, resample_fp, subsample_array,
                actor._raster.nodata, qi.dst_nodata,
                qi.interpolation, dst_array_slice,
            )
        else:
            self.dst_array_slice = dst_array_slice
            func = functools.partial(
                _resample_subsample_array,
                sample_fp, resample_fp, subsample_array,
                actor._raster.nodata, qi.dst_nodata,
                qi.interpolation, None,
            )
        actor._raster.debug_mngr.event('object_allocated', func)

        super().__init__(actor.address, func)

def _resample_subsample_array(sample_fp, resample_fp, subsample_array, src_nodata, dst_nodata, interpolation, dst_opt):
    """
    Parameters
    ----------
    sample_fp: Footprint of shape (Y, X)
        source footprint (before resampling)
    resample_fp: Footprint of shape (Y', X')
        destination footprint
    subsample_array: np.ndarray of shape (Y, X)
        source array (sould match sample_fp)
    src_nodata: None or nbr
    dst_nodata: nbr
    interpolation: str
    dst_opt: None or np.ndarray
        optional destination for resample

    Returns
    -------
    None or np.ndarray of shape (Y', X')
    """
    # TODO: Inplace remap
    res = ABackSourceRasterRemapMixin.remap(
        src_fp=sample_fp, dst_fp=resample_fp,
        array=subsample_array, mask=None,
        src_nodata=src_nodata, dst_nodata=dst_nodata,
        mask_mode='dilate', interpolation=interpolation,
    )
    if dst_opt is not None:
        dst_opt[:] = res
        return None
    else:
        return res



import numpy as np

def _slices_of_vector(vec):
    """Generates slices of oneline mask parts"""
    assert vec.ndim == 1
    diff = np.diff(np.r_[[False], vec, [False]].astype('int'))
    starts = np.where(diff == 1)[0]
    ends = np.where(diff == -1)[0]
    for s, e in zip(starts, ends):
        yield slice(s, e)

def slices_of_matrix(mask):
    """Generates slices of mask parts"""
    if mask is None:
        yield slice(0, None), slice(0, None)
        return

    ystart = None
    y = 0
    while True:
        # Iteration analysis
        if y == 0:
            begin_group = True
            send_group = False
            stop = False
        elif y == mask.shape[0]:
            begin_group = False
            send_group = True
            stop = True
        elif (mask[y - 1] != mask[y]).any():
            begin_group = True
            send_group = True
            stop = False
        else:
            begin_group = False
            send_group = False
            stop = False

        # Actions
        if send_group:
            yslice = slice(ystart, y)
            for xslice in _slices_of_vector(mask[ystart]):
                yield yslice, xslice
        if begin_group:
            ystart = y

        # Loop control
        if stop:
            break
        y += 1



"""Collection of spatial references manipulation utilities,

Will be filled/improved/moved in the future
"""

from osgeo import gdal, osr, ogr
import numpy as np
import affine

from buzzard._pint_interop import REG
from buzzard._env import env, Env
from buzzard._tools import conv
from buzzard import Footprint
from buzzard.srs._analysis import *

def wkt_of_any(string):
    """Wkt of user input"""
    out = osr.GetUserInputAsWKT(string)
    if isinstance(out, str):
        return out
    else:
        prj = None
        gdal_ds = gdal.OpenEx(string, conv.of_of_str('raster'))
        if gdal_ds is not None:
            prj = gdal_ds.GetProjection()
        gdal_ds = gdal.OpenEx(string, conv.of_of_str('vector'))
        if gdal_ds is not None:
            lyr = gdal_ds.GetLayerByIndex(0)
            if lyr is not None:
                prj = lyr.GetSpatialRef()
        if prj is not None:
            return prj.ExportToWkt()
    raise ValueError('Could not convert to wkt ({})'.format(str(gdal.GetLastErrorMsg()).strip('\n')))

def wkt_same(a, b):
    """Are two wkt equivalent"""
    if a == b:
        return True
    sra = osr.SpatialReference(a)
    srb = osr.SpatialReference(b)
    return bool(sra.IsSame(srb))

def _details_of_file(path):
    gdal_ds = gdal.OpenEx(path, conv.of_of_str('raster'))
    if gdal_ds is not None:
        aff = affine.Affine.from_gdal(*gdal_ds.GetGeoTransform())
        w, h = gdal_ds.RasterXSize, gdal_ds.RasterYSize
        cx, cy = aff * [w / 2, h / 2]
        return gdal_ds.GetProjection(), (cx, cy)
    gdal_ds = gdal.OpenEx(path, conv.of_of_str('vector'))
    if gdal_ds is not None:
        lyr = gdal_ds.GetLayerByIndex(0)
        if lyr is None:
            raise ValueError('Could not open file layer')
        extent = lyr.GetExtent()
        if extent is None:
            raise ValueError('Could not compute extent')
        minx, maxx, miny, maxy = extent
        cx, cy = (maxx + minx) / 2, (maxy + miny) / 2
        return lyr.GetSpatialRef().ExportToWkt(), (cx, cy)
    raise ValueError('Could not open file')

def wkt_of_file(path, center=False, unit=None, implicit_unit='m'):
    """Retrieve projection of file as wkt, optionally recenter projection, optionally change
    unit of projection. (Experimental function!)

    Parameters
    ----------
    path: str
    center: bool or nbr or (nbr, nbr)
    unit: None or str
    implicit_unit: str

    Returns
    -------
    str

    """
    wkt, centroid = _details_of_file(path)

    if center is False and unit is None:
        return wkt
    sr = osr.SpatialReference(wkt)

    if center is not False:
        if not sr.IsProjected():
            raise ValueError("Can't shift a spatial reference that is not projected")
        if center is True:
            center = (0, 0)
        else:
            center = np.asarray(center)
            if center.size == 1:
                center = (center, center)
            else:
                center = center.reshape(2)
        fe, fn = sr.GetProjParm('false_easting', 0.0), sr.GetProjParm('false_northing', 0.0)
        sr.SetProjParm('false_easting', fe - centroid[0] + center[0])
        sr.SetProjParm('false_northing', fn - centroid[1] + center[1])
        sr.SetAuthority('PROJCS', '', 0)

    if unit is not None:
        src_name = sr.GetLinearUnitsName()
        if src_name is None or src_name == '':
            src_name = implicit_unit
        src_u = REG.meter * sr.GetLinearUnits()
        dst_u = REG.parse_units(unit) * 1.0
        if dst_u.dimensionality != {'[length]': 1.0}:
            raise ValueError('todo')
        if src_u != dst_u:
            sr.SetLinearUnitsAndUpdateParameters(unit, float(dst_u / REG.m))
        sr.SetAuthority('PROJCS', '', 0)
    return sr.ExportToPrettyWkt()



# pylint: disable=redefined-outer-name, unused-argument

from __future__ import division, print_function
import os
import tempfile
import uuid
import string
import weakref
import gc

import numpy as np
from osgeo import gdal
import pytest
import shapely.ops

import buzzard as buzz
from buzzard.test.tools import fpeq, sreq, eq
from buzzard.test import make_tile_set
from .tools import  get_srs_by_name

SR1 = get_srs_by_name('EPSG:2154')
SR2 = get_srs_by_name('EPSG:2154 centered')

# FIXTURES ************************************************************************************** **
@pytest.fixture(scope='module')
def fps():
    """See make_tile_set
    A B C
    D E F
    G H I
    """
    return make_tile_set.make_tile_set(3, [0.1, -0.1])

@pytest.fixture()
def random_path_tif():
    """Create a temporary path, and take care of cleanup afterward"""
    path = '{}/{}.tif'.format(tempfile.gettempdir(), uuid.uuid4())
    yield path
    if os.path.isfile(path):
        try:
            gdal.GetDriverByName('GTiff').Delete(path)
        except:
            os.remove(path)

@pytest.fixture()
def random_path_shp():
    """Create a temporary path, and take care of cleanup afterward"""
    path = '{}/{}.shp'.format(tempfile.gettempdir(), uuid.uuid4())
    yield path
    if os.path.isfile(path):
        try:
            gdal.GetDriverByName('ESRI Shapefile').Delete(path)
        except:
            os.remove(path)

def test_raster(fps, random_path_tif):

    def _asserts(should_exist, should_be_open, is_anonymous=False):

        exist = os.path.isfile(random_path_tif)
        assert should_exist == exist

        if not is_anonymous:
            is_open_key = 'test' in ds
            is_open_prox = test in ds
            assert is_open_key == is_open_prox
            assert is_open_key == should_be_open
            if is_open_key:
                assert test is ds.test is ds['test']
        else:
            is_open = test in ds
            assert is_open == should_be_open

        if should_be_open and not is_anonymous:
            assert ds['test'] == ds.test
        if should_be_open:
            assert len(ds) == 1
        else:
            assert len(ds) == 0

    ds = buzz.Dataset()

    # Raster test 1
    test = None
    _asserts(False, False)
    test = ds.create_raster('test', random_path_tif, fps.A, float, 1)
    _asserts(True, True)
    ds.test.close()
    _asserts(True, False)

    test = ds.aopen_raster(random_path_tif)
    _asserts(True, True, True)
    test.close()
    _asserts(True, False)

    test = ds.open_raster('test', random_path_tif, mode='w')
    _asserts(True, True)
    test.remove()
    _asserts(False, False)

    # Raster test 2 - context/close
    with ds.create_raster('test', random_path_tif, fps.A, float, 1).close as test:
        _asserts(True, True)
    _asserts(True, False)
    with ds.open_raster('test', random_path_tif, mode='w').delete as test:
        _asserts(True, True)
    _asserts(False, False)

    # Raster test 3 - context/close/anonymous
    with ds.acreate_raster(random_path_tif, fps.A, float, 1).delete as test:
        _asserts(True, True, True)
    _asserts(False, False)

    # Raster test 4 - context/delete
    with ds.create_raster('test', random_path_tif, fps.A, float, 1).delete as test:
        _asserts(True, True)
    _asserts(False, False)

    # Raster test 5 - MEM
    with ds.create_raster('test', '', fps.A, float, 1, driver='MEM').close as test:
        _asserts(False, True)

    # Raster test 6 - numpy
    with ds.wrap_numpy_raster('test', fps.A, np.zeros(fps.A.shape)).close as test:
        _asserts(False, True)

    # Raster test 7 - gc
    del ds
    ws = weakref.WeakSet()
    def _test():
        ds = buzz.Dataset()
        assert len(ds) == len(ws) == 0

        prox = ds.create_raster('test', random_path_tif, fps.A, float, 1)
        ws.add(prox)
        assert len(ds) == len(ws) == 1

        prox = ds.acreate_raster('', fps.A, float, 1, driver='MEM')
        ws.add(prox)
        assert len(ds) == len(ws) == 2

        prox = ds.awrap_numpy_raster(fps.A, np.zeros(fps.A.shape))
        ws.add(prox)
        assert len(ds) == len(ws) == 3

        ws.add(ds)
        assert len(ws) == 4

    _test()
    gc.collect()
    assert len(ws) == 0


def test_vector(fps, random_path_shp):

    def _asserts(should_exist, should_be_open, is_anonymous=False):

        exist = os.path.isfile(random_path_shp)
        assert should_exist == exist

        if not is_anonymous:
            is_open_key = 'test' in ds
            is_open_prox = test in ds
            assert is_open_key == is_open_prox
            assert is_open_key == should_be_open
            if is_open_key:
                assert test is ds.test is ds['test']
        else:
            is_open = test in ds
            assert is_open == should_be_open

        if should_be_open and not is_anonymous:
            assert ds['test'] == ds.test
        if should_be_open:
            assert len(ds) == 1
        else:
            assert len(ds) == 0

    ds = buzz.Dataset()

    # Vector test 1
    test = None
    _asserts(False, False)
    test = ds.create_vector('test', random_path_shp, 'polygon')
    _asserts(True, True)
    ds.test.close()
    _asserts(True, False)

    test = ds.aopen_vector(random_path_shp)
    _asserts(True, True, True)
    test.close()
    _asserts(True, False)

    test = ds.open_vector('test', random_path_shp, mode='w')
    _asserts(True, True)
    test.remove()
    _asserts(False, False)

    # Vector test 2 - context/close
    with ds.create_vector('test', random_path_shp, 'polygon').close as test:
        _asserts(True, True)
    _asserts(True, False)
    with ds.open_vector('test', random_path_shp, mode='w').delete as test:
        _asserts(True, True)
    _asserts(False, False)

    # Vector test 3 - context/close/anonymous
    with ds.acreate_vector(random_path_shp, 'polygon').delete as test:
        _asserts(True, True, True)
    _asserts(False, False)

    # Vector test 4 - context/delete
    with ds.create_vector('test', random_path_shp, 'polygon').delete as test:
        _asserts(True, True)
    _asserts(False, False)

    # Vector test 5 - MEM
    with ds.create_vector('test', '', 'polygon', driver='Memory').close as test:
        _asserts(False, True)

    # Vector test 6 - gc
    del ds
    ws = weakref.WeakSet()
    def _test():
        ds = buzz.Dataset()
        assert len(ds) == len(ws) == 0

        prox = ds.create_vector('test', random_path_shp, 'polygon')
        ws.add(prox)
        assert len(ds) == len(ws) == 1

        prox = ds.acreate_vector('', 'polygon', driver='Memory')
        ws.add(prox)
        assert len(ds) == len(ws) == 2

        ws.add(ds)
        assert len(ws) == 3

    _test()
    gc.collect()
    assert len(ws) == 0



import contextlib
import os

import numpy as np
from osgeo import gdal, ogr, osr
import shapely
import shapely.ops
import shapely.geometry as sg

from buzzard._a_stored_vector import ABackStoredVector
from buzzard._tools import conv, GDALErrorCatcher
from buzzard._env import Env

class ABackGDALVector(ABackStoredVector):
    """Abstract class defining the common implementation of all vector formats in OGR"""

    # extent/len implementation ***************************************************************** **
    @property
    def extent(self):
        with self.acquire_driver_object() as (_, lyr):
            extent = lyr.GetExtent()

        if extent is None: # pragma: no cover
            raise ValueError('Could not compute extent')
        if self.to_work:
            xa, xb, ya, yb = extent
            extent = self.to_work([[xa, ya], [xb, yb]])
            extent = np.asarray(extent)[:, :2]
            extent = extent[0, 0], extent[1, 0], extent[0, 1], extent[1, 1]
        return np.asarray(extent)

    @property
    def extent_stored(self):
        with self.acquire_driver_object() as (_, lyr):
            extent = lyr.GetExtent()
        if extent is None: # pragma: no cover
            raise ValueError('Could not compute extent')
        return np.asarray(extent)

    def __len__(self):
        """Return the number of features in vector layer"""
        with self.acquire_driver_object() as (_, lyr):
            return len(lyr)

    # iter_data implementation ****************************************************************** **
    def iter_data(self, geom_type, field_indices, slicing, mask_poly, mask_rect, clip):
        clip_poly = None
        if mask_poly is not None:
            mask_poly = conv.ogr_of_shapely(mask_poly)
            if clip:
                clip_poly = mask_poly
        elif mask_rect is not None:
            if clip:
                clip_poly = conv.ogr_of_shapely(sg.box(*mask_rect))

        ftr = None # Necessary to prevent the old swig bug
        geom = None # Necessary to prevent the old swig bug
        with self.acquire_driver_object() as (_, lyr):
            for ftr in self.iter_features_driver(slicing, mask_poly, mask_rect, lyr):
                geom = ftr.geometry()
                if geom is None or geom.IsEmpty():
                    # `geom is None` and `geom.IsEmpty()` is not exactly the same case, but whatever
                    geom = None
                    if not self.back_ds.allow_none_geometry: # pragma: no cover
                        raise Exception(
                            'None geometry in feature '
                            '(Set `allow_none_geometry=True` in Dataset constructor to silence)'
                        )
                else:
                    if clip:
                        geom = geom.Intersection(clip_poly)
                        assert not geom.IsEmpty()
                    geom = conv.shapely_of_ogr(geom)
                    if self.to_work:
                        geom = shapely.ops.transform(self.to_work, geom)
                    if geom_type == 'coordinates':
                        geom = sg.mapping(geom)['coordinates']
                    elif geom_type == 'geojson':
                        geom = sg.mapping(geom)

                yield (geom,) + tuple(
                    self._type_of_field_index[index](ftr.GetField(index))
                    if ftr.GetField(index) is not None
                    else None
                    for index in field_indices
                )

        # Necessary to prevent the old swig bug
        # https://trac.osgeo.org/gdal/ticket/6749
        del geom
        del ftr
        del clip_poly
        del mask_rect, mask_poly

    @staticmethod
    def iter_features_driver(slicing, mask_poly, mask_rect, lyr):
        with contextlib.ExitStack() as stack:
            stack.push(lambda *args, **kwargs: lyr.ResetReading())
            if mask_poly is not None:
                lyr.SetSpatialFilter(mask_poly)
                stack.push(lambda *args, **kwargs: lyr.SetSpatialFilter(None))
            elif mask_rect is not None:
                lyr.SetSpatialFilterRect(*mask_rect)
                stack.push(lambda *args, **kwargs: lyr.SetSpatialFilter(None))

            start, stop, step = slicing.indices(len(lyr))
            indices = range(start, stop, step)
            ftr = None # Necessary to prevent the old swig bug
            if step == 1:
                lyr.SetNextByIndex(start)
                for i in indices:
                    ftr = lyr.GetNextFeature()
                    if ftr is None: # pragma: no cover
                        raise IndexError('Feature #{} not found'.format(i))
                    yield ftr
            else:
                for i in indices:
                    lyr.SetNextByIndex(i)
                    ftr = lyr.GetNextFeature()
                    if ftr is None: # pragma: no cover
                        raise IndexError('Feature #{} not found'.format(i))
                    yield ftr

        # Necessary to prevent the old swig bug
        # https://trac.osgeo.org/gdal/ticket/6749
        del slicing, mask_poly, mask_rect, ftr

    # insert_data implementation **************************************************************** **
    def insert_data(self, geom, geom_type, fields, index):
        geom = self._ogr_of_geom(geom, geom_type)
        with self.acquire_driver_object() as (_, lyr):
            ftr = ogr.Feature(lyr.GetLayerDefn())

            if geom is not None:
                err = ftr.SetGeometry(geom)
                if err: # pragma: no cover
                    raise ValueError('Could not set geometry (%s)' % str(gdal.GetLastErrorMsg()).strip('\n'))

                if not self.back_ds.allow_none_geometry and ftr.GetGeometryRef() is None: # pragma: no cover
                    raise ValueError(
                        'Invalid geometry inserted '
                        '(Set `allow_none_geometry=True` in Dataset constructor to silence)'
                    )

            if index >= 0:
                err = ftr.SetFID(index)
                if err: # pragma: no cover
                    raise ValueError('Could not set field id (%s)' % str(gdal.GetLastErrorMsg()).strip('\n'))
            for i, field in enumerate(fields):
                if field is not None:
                    err = ftr.SetField2(i, self._type_of_field_index[i](field))
                    if err: # pragma: no cover
                        raise ValueError('Could not set field #{} ({}) ({})'.format(
                            i, field, str(gdal.GetLastErrorMsg()).strip('\n')
                        ))
            passed = ftr.Validate(ogr.F_VAL_ALL, True)
            if not passed: # pragma: no cover
                raise ValueError('Invalid feature {} ({})'.format(
                    err, str(gdal.GetLastErrorMsg()).strip('\n')
                ))

            err = lyr.CreateFeature(ftr)
            if err: # pragma: no cover
                raise ValueError('Could not create feature {} ({})'.format(
                    err, str(gdal.GetLastErrorMsg()).strip('\n')
                ))

    def _ogr_of_geom(self, geom, geom_type):
        if geom_type is None:
            geom = geom
        elif self.to_virtual:
            if geom_type == 'coordinates':
                geom = sg.shape({
                    'type': self.type,
                    'coordinates': geom,
                })
            geom = shapely.ops.transform(self.to_virtual, geom)
            # geom = conv.ogr_of_shapely(geom)
            # TODO: Use json and unit test
            mapping = sg.mapping(geom)
            geom = conv.ogr_of_coordinates(
                mapping['coordinates'],
                mapping['type'],
            )
            if geom is None: # pragma: no cover
                raise ValueError('Could not convert `{}` of type `{}` to `ogr.Geometry`'.format(
                    geom_type, self.type
                ))
        elif geom_type == 'coordinates':
            geom = conv.ogr_of_coordinates(geom, self.type)
            if geom is None: # pragma: no cover
                raise ValueError('Could not convert `{}` of type `{}` to `ogr.Geometry`'.format(
                    geom_type, self.type
                ))
        elif geom_type == 'shapely':
            # geom = conv.ogr_of_shapely(geom)
            # TODO: Use json and unit test
            mapping = sg.mapping(geom)
            geom = conv.ogr_of_coordinates(
                mapping['coordinates'],
                mapping['type'],
            )
            if geom is None: # pragma: no cover
                raise ValueError('Could not convert `{}` of type `{}` to `ogr.Geometry`'.format(
                    geom_type, self.type
                ))
        else:
            assert False # pragma: no cover
        return geom

    # Misc ************************************************************************************** **
    def acquire_driver_object(self): # pragma: no cover
        raise NotImplementedError('ABackGDALRaster.acquire_driver_object is virtual pure')

    @classmethod
    def create_file(cls, path, geometry, fields, layer, driver, options, wkt, ow):
        """Create a vector dataset"""

        # Step 0 - Find driver ********************************************** **
        success, payload = GDALErrorCatcher(gdal.GetDriverByName, none_is_error=True)(driver)
        if not success:
            raise ValueError('Could not find a driver named `{}` (gdal error: `{}`)'.format(
                driver, payload[1]
            ))
        dr = payload

        # Step 1 - Overwrite ************************************************ **
        if dr.ShortName != 'Memory' and os.path.exists(path):
            if ow:
                success, payload = GDALErrorCatcher(dr.Delete, nonzero_int_is_error=True)(path)
                if not success:
                    msg = 'Could not delete `{}` using driver `{}` (gdal error: `{}`)'.format(
                        path, dr.ShortName, payload[1]
                    )
                    raise RuntimeError(msg)
            else:
                msg = "Can't create `{}` with `ow=False` (overwrite) because file exist".format(
                    path,
                )
                raise RuntimeError(msg)

        # Step 2 - Create gdal_ds ******************************************* **
        success, payload = GDALErrorCatcher(dr.Create)(path, 0, 0, 0, 0, options)
        if not success: # pragma: no cover
            raise RuntimeError('Could not create `{}` using driver `{}` (gdal error: `{}`)'.format(
                path, dr.ShortName, payload[1]
            ))
        gdal_ds = payload

        # Step 3 - Get spatial reference ************************************ **
        if wkt is not None:
            sr = osr.SpatialReference(wkt)
        else:
            sr = None

        # Step 4 - Create layer ********************************************* **
        geometry = conv.wkbgeom_of_str(geometry)
        success, payload = GDALErrorCatcher(gdal_ds.CreateLayer, none_is_error=True)(
            layer, sr, geometry, options
        )
        if not success: # pragma: no cover
            msg = 'Could not create layer `{}` in `{}` using driver `{}` (gdal error: `{}`)'.format(
                layer, path, dr.ShortName, payload[1]
            )
            raise RuntimeError(msg)
        lyr = payload

        # Step 5 - Set fields *********************************************** **
        for field in fields:
            flddef = ogr.FieldDefn(field['name'], field['type'])
            if field['precision'] is not None:
                flddef.SetPrecision(field['precision'])
            if field['width'] is not None:
                flddef.SetWidth(field['width'])
            if field['nullable'] is not None:
                flddef.SetNullable(field['nullable'])
            if field['default'] is not None:
                flddef.SetDefault(field['default'])
            lyr.CreateField(flddef)
        lyr.SyncToDisk()
        gdal_ds.FlushCache()

        return gdal_ds, lyr

    @classmethod
    def _fields_of_lyr(cls, lyr):
        """Used on file opening / creation"""
        featdef = lyr.GetLayerDefn()
        field_count = featdef.GetFieldCount()
        return [cls._field_of_def(featdef.GetFieldDefn(i)) for i in range(field_count)]

    @staticmethod
    def _field_of_def(fielddef):
        """Used on file opening / creation"""
        oft = fielddef.type
        oftstr = conv.str_of_oft(oft)
        type_ = conv.type_of_oftstr(oftstr)
        default = fielddef.GetDefault()
        return {
            'name': fielddef.name,
            'precision': fielddef.precision,
            'width': fielddef.width,
            'nullable': bool(fielddef.IsNullable()),
            'default': None if default is None else type_(default),
            'type': oftstr,
        }



""">>> help(CallOrContext)
>>> help(Singleton)
"""

from osgeo import gdal

from buzzard._tools import conv

class GDALErrorCatcher:
    """Wrap a call to a gdal/ogr/osr function to streamline the behavior of gdal no matter the
    type of function and no matter the global states:
    - `gdal.*UseException` functions modify global states,
    - `gdal.*ErrorHandler` functions modify thread-local states.

    Using this wrapper makes gdal errors thread-safe.

    A related update was included in `gdal>=2.3.3` (https://github.com/OSGeo/gdal/pull/1117) but
    the python bindings did not change. To be safe, buzzard should constrain `gdal>=2.3.3`.

    ### Type of gdal functions
    - Return `obj` on success, `None` on failure but trigger `error_handler`
      - `Driver.Create`
    - Return `obj` on success, `None` on failure but does not trigger `error_handler`
      - `gdal.GetDriverByName`
      - `gdal.OpenEx`
    - Return `error code`, but does not trigger `error_handler`
      - `Driver.Delete`
    - Return `obj` on success, `error code` on failure, but does not trigger `error_handler`
      - `osr.GetUserInputAsWKT`

    Examples
    --------
    >>> gdal.UseExceptions()
    ... dr = gdal.GetDriverByName('mem')
    ... success, payload = GDALErrorCatcher(dr.Create)('', 0, 0)
    ... print(success, payload)
    False (3, 1, 'Attempt to create 0x0 dataset is illegal,sizes must be larger than zero.') # no exception raised

    >>> gdal.DontUseExceptions()
    ... dr = gdal.GetDriverByName('mem')
    ... success, payload = GDALErrorCatcher(dr.Create)('', 0, 0)
    ... print(success, payload)
    False (3, 1, 'Attempt to create 0x0 dataset is illegal,sizes must be larger than zero.')

    >>> dr = gdal.GetDriverByName('mem')
    ... success, payload = GDALErrorCatcher(dr.Create)('', 1, 1)
    ... print(success, payload)
    True <osgeo.gdal.Dataset; proxy of <Swig Object of type 'GDALDatasetShadow *' at 0x> >

    """
    def __init__(self, fn, nonzero_int_is_error=False, none_is_error=False):
        self._fn = fn
        self._nonzero_int_is_error = nonzero_int_is_error
        self._none_is_error = none_is_error

    def __call__(self, *args, **kwargs):
        errs, res = None, None

        def error_handler(err_level, err_no, err_msg):
            nonlocal errs
            if err_level >= gdal.CE_Failure:
                errs = err_no, err_msg

        gdal.PushErrorHandler(error_handler)
        try:
            res = self._fn(*args, **kwargs)
        except Exception as e:
            if errs is None:
                # This is not a gdal error, this might be a swig error or something else
                raise
            else:
                # This is a gdal error, and `gdal.GetUseExceptions()` is True
                # Read problems from the `errs` variable, the details stored in this exception
                # are not reliable.
                pass
        finally:
            gdal.PopErrorHandler()

        if errs:
            return False, errs
        if self._none_is_error and res is None:
            return False, (0, str(gdal.GetLastErrorMsg()).strip('\n'))
        if self._nonzero_int_is_error and isinstance(res, int) and res != 0:
            return False, (res, conv.str_of_cple(res))
        return True, res

class CallOrContext(object):
    """Private helper class to provide a common behaviour both on call and on exit"""
    def __init__(self, obj, routine):
        self._obj = obj
        self._routine = routine

    def __call__(self):
        self._routine()

    def __enter__(self):
        return self._obj

    def __exit__(self, *args, **kwargs):
        self._routine()

class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called.
    works in Python 2 & 3
    https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class Singleton(_Singleton('SingletonMeta', (object,), {})):
    pass

class _Any(object):
    """Helper for pattern matching"""
    def __eq__(self, _):
        return True

ANY = _Any()



import uuid
import contextlib

from osgeo import gdal
import numpy as np

from buzzard._a_pooled_emissary_vector import APooledEmissaryVector, ABackPooledEmissaryVector
from buzzard._a_gdal_vector import ABackGDALVector
from buzzard._tools import conv, GDALErrorCatcher

class GDALFileVector(APooledEmissaryVector):
    """Concrete class defining the behavior of a GDAL vector using a file

    >>> help(Dataset.open_vector)
    >>> help(Dataset.create_vector)

    Features Defined
    ----------------
    None
    """

    def __init__(self, ds, allocator, open_options, mode):
        back = BackGDALFileVector(
            ds._back, allocator, open_options, mode,
        )
        super(GDALFileVector, self).__init__(ds=ds, back=back)

class BackGDALFileVector(ABackPooledEmissaryVector, ABackGDALVector):
    """Implementation of GDALFileVector"""

    def __init__(self, back_ds, allocator, open_options, mode):
        uid = uuid.uuid4()

        with back_ds.acquire_driver_object(uid, allocator) as (gdal_ds, lyr):
            rect = None
            if lyr is not None:
                rect = lyr.GetExtent()
            path = gdal_ds.GetDescription()
            driver = gdal_ds.GetDriver().ShortName
            sr = lyr.GetSpatialRef()
            if sr is None:
                wkt_stored = None
            else:
                wkt_stored = sr.ExportToWkt()
            fields = BackGDALFileVector._fields_of_lyr(lyr)
            type = conv.str_of_wkbgeom(lyr.GetGeomType())
            layer = lyr.GetName()

        super(BackGDALFileVector, self).__init__(
            back_ds=back_ds,
            wkt_stored=wkt_stored,
            mode=mode,
            driver=driver,
            open_options=open_options,
            path=path,
            uid=uid,
            layer=layer,
            fields=fields,
            rect=rect,
            type=type
        )

        self._type_of_field_index = [
            conv.type_of_oftstr(field['type'])
            for field in self.fields
        ]

    def allocator(self):
        return self.open_file(self.path, self.layer, self.driver, self.open_options, self.mode)

    @staticmethod
    def open_file(path, layer, driver, options, mode):
        """Open a vector dataset"""
        options = [str(arg) for arg in options] if len(options) else []
        success, payload = GDALErrorCatcher(gdal.OpenEx, none_is_error=True)(
            path,
            conv.of_of_mode(mode) | conv.of_of_str('vector'),
            [driver],
            options,
        )
        if not success:
            raise RuntimeError('Could not open `{}` using driver `{}` (gdal error: `{}`)'.format(
                path, driver, payload[1]
            ))
        gdal_ds = payload

        if layer is None:
            layer = 0
        if np.all(np.isreal(layer)):
            success, payload = GDALErrorCatcher(gdal_ds.GetLayer)(int(layer))
        else:
            success, payload = GDALErrorCatcher(gdal_ds.GetLayerByName)(layer)

        if not success: # pragma: no cover
            count = gdal_ds.GetLayerCount()
            raise Exception('Could not open layer `{}` of `{}` ({} layers available: {}) (gdal error: `{}`)'.format(
                layer,
                path,
                count,
                {i: gdal_ds.GetLayerByIndex(i).GetName() for i in range(count)},
                payload[1],
            ))
        lyr = payload

        return gdal_ds, lyr

    @contextlib.contextmanager
    def acquire_driver_object(self):
        with self.back_ds.acquire_driver_object(
            self.uid,
            self.allocator,
        ) as gdal_objs:
            yield gdal_objs

    def delete(self):
        super(BackGDALFileVector, self).delete()

        success, payload = GDALErrorCatcher(gdal.GetDriverByName, none_is_error=True)(self.driver)
        if not success: # pragma: no cover
            raise ValueError('Could not find a driver named `{}` (gdal error: `{}`)'.format(
                self.driver, payload[1]
            ))
        dr = payload

        success, payload = GDALErrorCatcher(dr.Delete, nonzero_int_is_error=True)(self.path)
        if not success: # pragma: no cover
            raise RuntimeError('Could not delete `{}` using driver `{}` (gdal error: `{}`)'.format(
                self.path, dr.ShortName, payload[1]
            ))



"""Test:
- Vector opening
- Vector creation
- Attributes validity
- Vector deletion
"""

# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import itertools
import os
import uuid
import tempfile
import operator
from pprint import pprint

import numpy as np
import pytest
from osgeo import gdal
import shapely.geometry as sg

from .tools import get_srs_by_name, eq
from buzzard import Footprint, Dataset, srs

# SR1 = get_srs_by_name('EPSG:2154')
SR1 = dict(wkt="""GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.01745329251994328,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]""")

DRIVER_LIST = [
    'ESRI Shapefile',
    'GeoJson',
    'DGN',
    'DXF',
    'Memory',
]

DRIVER_STORES_SRS = {
    'ESRI Shapefile': True,
    'GeoJson': False,
    'DGN': False,
    'DXF': False,
    'Memory': True,
}

DRIVER_STORES_GTYPE = {
    'ESRI Shapefile': True,
    'GeoJson': True,
    'DGN': False,
    'DXF': False,
    'Memory': True,
}

DRIVER_SUPPORT_EMPTY_COLLECTION = {
    'ESRI Shapefile': True,
    'GeoJson': True,
    'DGN': False,
    'DXF': False,
    'Memory': True,
}

EXTENSION_OF_DRIVER = {
    'ESRI Shapefile': '.shp',
    'GeoJson': '.json',
    'DGN': '.dgn',
    'DXF': '.dxf',
}

GTYPES_OF_DRIVER = {
    'ESRI Shapefile': {
        'GeometryCollection',
        'GeometryCollection25D',
        'LinearRing',
        'LineString',
        'LineString25D',
        'MultiLineString',
        'MultiLineString25D',
        'MultiPoint',
        'MultiPoint25D',
        'MultiPolygon',
        'MultiPolygon25D',
        'Point',
        'Point25D',
        'Polygon',
        'Polygon25D',
    },

    'GeoJson': {
        'GeometryCollection',
        'GeometryCollection25D',
        'LinearRing',
        'LineString',
        'LineString25D',
        'MultiLineString',
        'MultiLineString25D',
        'MultiPoint',
        'MultiPoint25D',
        'MultiPolygon',
        'MultiPolygon25D',
        'Point',
        'Point25D',
        'Polygon',
        'Polygon25D',
    },

    'DGN': {
        'GeometryCollection',
        # 'GeometryCollection25D',
        'LinearRing',
        'LineString',
        # 'LineString25D',
        'MultiLineString',
        # 'MultiLineString25D',
        # 'MultiPoint',
        # 'MultiPoint25D',
        # 'MultiPolygon',
        # 'MultiPolygon25D',
        # 'Point',
        # 'Point25D',
        # 'Polygon',
        # 'Polygon25D',
    },

    'DXF': {
        'GeometryCollection',
        'GeometryCollection25D',
        'LinearRing',
        'LineString',
        'LineString25D',
        'MultiLineString',
        'MultiLineString25D',
        # 'MultiPoint',
        # 'MultiPoint25D',
        'MultiPolygon',
        'MultiPolygon25D',
        'Point',
        'Point25D',
        'Polygon',
        'Polygon25D',
    },

    'Memory': {
        'GeometryCollection',
        'GeometryCollection25D',
        'LinearRing',
        'LineString',
        'LineString25D',
        'MultiLineString',
        'MultiLineString25D',
        'MultiPoint',
        'MultiPoint25D',
        'MultiPolygon',
        'MultiPolygon25D',
        'Point',
        'Point25D',
        'Polygon',
        'Polygon25D',
    },
}


FTYPES_OF_DRIVER = {
    'ESRI Shapefile': {
       'integer64',
       'real',
       'string',
       # 'integer64list',
       # 'reallist',
       # 'stringlist',
    },

    'GeoJson': {
       'integer64',
       'real',
       'string',
       'integer64list',
       'reallist',
       'stringlist',
    },

    'DGN': {
       # 'integer64',
       # 'real',
       # 'string',
       # 'integer64list',
       # 'reallist',
       # 'stringlist',
    },

    'DXF': {
       # 'integer64',
       # 'real',
       # 'string',
       # 'integer64list',
       # 'reallist',
       # 'stringlist',
    },

    'Memory': {
       'integer64',
       'real',
       'string',
       'integer64list',
       'reallist',
       'stringlist',
    },

}

pt0 = sg.Point(0, 0)
mpt0 = sg.MultiPoint([])
mpt1 = sg.MultiPoint([pt0])

ptz0 = sg.Point(0, 0, 0)
mptz0 = sg.MultiPoint([])
mptz1 = sg.MultiPoint([ptz0])

ls0 = sg.LineString([[0, 0], [1, 1]])
mls0 = sg.MultiLineString([])
mls1 = sg.MultiLineString([ls0])

lsz0 = sg.LineString([[0, 0, 0], [1, 1, 1]])
mlsz0 = sg.MultiLineString([])
mlsz1 = sg.MultiLineString([lsz0])

p0 = sg.Polygon(
	([13, 10], [10, 10], [10, 13], [13, 13], [13, 10]),
	[([12, 11], [12, 12], [11, 12], [11, 11], [12, 11])],
)
mp0 = sg.MultiPolygon([])
mp1 = sg.MultiPolygon([p0])

pz0 = sg.Polygon(
	([13, 10, 42], [10, 10, 42], [10, 13, 42], [13, 13, 42], [13, 10, 42]),
	[([12, 11, 42], [12, 12, 42], [11, 12, 42], [11, 11, 42], [12, 11, 42])],
)
mpz0 = sg.MultiPolygon([])
mpz1 = sg.MultiPolygon([pz0])

GEOM_TESTS = [
    ['Point', []],
    ['Point', [pt0]],
    ['MultiPoint', []],
    ['MultiPoint', [mpt0, mpt1]],
    ['Point25D', []],
    ['Point25D', [ptz0]],
    ['MultiPoint25D', [mptz0, mptz1]],

	['LineString', []],
	['LineString', [ls0]],
	['MultiLineString', []],
	['MultiLineString', [mls0, mls1]],
	['LineString25D', []],
	['LineString25D', [lsz0]],
	['MultiLineString25D', []],
	['MultiLineString25D', [mlsz0, mlsz1]],

    ['Polygon', []],
    ['Polygon', [p0]],
    ['MultiPolygon', []],
    ['MultiPolygon', [mp0, mp1]],
    ['Polygon25D', []],
    ['Polygon25D', [pz0]],
    ['MultiPolygon25D', []],
    ['MultiPolygon25D', [mpz0, mpz1]],

    # 'GeometryCollection', 'GeometryCollection25D', 'LinearRing', 'LineString25D', 'MultiLineString25D', 'Point25D', 'MultiPoint25D', 'Polygon25D', 'MultiPolygon25D', 'LineString', 'MultiLineString', 'Point', 'MultiPoint', 'Polygon', 'MultiPolygon'
# ])
]

FIELD_TESTS = [[],
               [
    [dict(name='i', type=int), 42],
    [dict(name='f', type=float), 42.42],
    [dict(name='s', type=str), 'ft'],
    [dict(name='is1', type='int list'), ()],
    [dict(name='is2', type='int list'), (4, 2)],
    [dict(name='fs1', type='float list'), ()],
    [dict(name='fs2', type='float list'), (4.2, 4.2)],
    [dict(name='ss1', type='str list'), ()],
    [dict(name='ss2', type='str list'), ('f', 't')],
]
]

def pytest_generate_tests(metafunc):
    tests = []

    if 'driver_file' in metafunc.fixturenames:
        driver_list = [d for d in DRIVER_LIST if d != 'Memory']
    else:
        driver_list = [d for d in DRIVER_LIST if d == 'Memory']

    for driver in driver_list:
        for gtype, geoms in GEOM_TESTS:
            if gtype not in GTYPES_OF_DRIVER[driver]:
                continue

            mask = [
                not g.is_empty or DRIVER_SUPPORT_EMPTY_COLLECTION[driver]
                for g in geoms
            ]
            geoms = list(itertools.compress(geoms, mask))

            for field_test in FIELD_TESTS:
                ftype = list(map(operator.itemgetter(0), field_test))
                fields = list(map(operator.itemgetter(1), field_test))
                mask = [
                    d['type'] in FTYPES_OF_DRIVER[driver]
                    for d in ftype
                ]
                ftype = list(itertools.compress(ftype, mask))
                fields = list(itertools.compress(fields, mask))
                tests.append((
                    driver, gtype, geoms, ftype, fields
                ))
    if 'driver_file' in metafunc.fixturenames:
        metafunc.parametrize(
            argnames='driver_file,gtype,geoms,ftypes,fields',
            argvalues=tests,
        )
    else:
        metafunc.parametrize(
            argnames='driver_mem,gtype,geoms,ftypes,fields',
            argvalues=tests,
        )

@pytest.fixture
def path(driver_file):
    ext = EXTENSION_OF_DRIVER[driver_file]
    path = '{}/{}{}'.format(tempfile.gettempdir(), uuid.uuid4(), ext)
    yield path
    if os.path.isfile(path):
        try:
            gdal.GetDriverByName(driver_file).Delete(path)
        except:
            os.remove(path)

def test_file(path, driver_file, gtype, geoms, ftypes, fields):
    driver = driver_file
    ds = Dataset(allow_none_geometry=1)

    with ds.acreate_vector(path, gtype, ftypes, driver=driver, options=[], sr=SR1['wkt']).close as v:
        for geom in geoms:
            v.insert_data(geom, fields)

        # TESTS 0
        if DRIVER_STORES_SRS[driver]:
            assert srs.wkt_same(v.wkt_stored, SR1['wkt'])
            assert srs.wkt_same(v.wkt_virtual, SR1['wkt'])
        for ftype, ftype_stored in zip(ftypes, v.fields):
            for key, value in ftype.items():
                assert ftype_stored[key] == value
        assert v.mode == 'w'
        if DRIVER_STORES_GTYPE[driver]:
            assert v.type.lower() == gtype.lower()

    assert os.path.isfile(path)
    with ds.aopen_vector(path, driver=driver, options=[]).close as v:

        # TESTS 0
        if DRIVER_STORES_SRS[driver]:
            assert srs.wkt_same(v.wkt_stored, SR1['wkt'])
            assert srs.wkt_same(v.wkt_virtual, SR1['wkt'])
        for ftype, ftype_stored in zip(ftypes, v.fields):
            for key, value in ftype.items():
                assert ftype_stored[key] == value
        assert v.mode == 'r'
        # if DRIVER_STORES_GTYPE[driver]:
        #     assert v.type.lower() == gtype.lower()

        # TESTS 1
        datas = list(v.iter_data())
        assert len(datas) == len(geoms)
        assert len(v) == len(geoms)
        # assert v.layer in {0, ''}

        if not sg.GeometryCollection(geoms).is_empty:
            assert eq(
                v.bounds,
                v.bounds_stored,
                v.extent[[0, 2, 1, 3]],
                v.extent_stored[[0, 2, 1, 3]],
                sg.GeometryCollection(geoms).bounds
            )

        for geom, data in zip(geoms, datas):
            if not isinstance(data, tuple):
                data = (data,)
            read_geom, *read_fields = data
            if read_geom is None or read_geom.is_empty:
                assert geom.is_empty
            else:
                assert (geom ^ read_geom).is_empty, (
                    geom.wkt,
                    read_geom.wkt,
                )
        with pytest.raises(RuntimeError):
            v.delete()

    # Test error - driver name
    with pytest.raises(ValueError, match='driver'):
        ds.acreate_vector(path, gtype, ftypes, driver='Truc42', options=[], sr=SR1['wkt'])

    # Test error - need overwrite
    with pytest.raises(RuntimeError, match='overwrite'):
        ds.acreate_vector(path, gtype, ftypes, driver=driver, options=[], sr=SR1['wkt'], ow=False)

    # Test deletion
    with ds.aopen_vector(path, driver=driver, options=[], mode='w').delete as v:
        assert v.mode == 'w'
    assert not os.path.exists(path)

    # Test error - open
    with pytest.raises(RuntimeError, match='Could not open'):
        ds.aopen_vector(path, driver=driver)

    # Test error - bad sr
    with pytest.raises(ValueError, match='wkt'):
        ds.acreate_vector(path, gtype, ftypes, driver=driver, options=[], sr="ca n'existe pas")

    # Test error - failed file deletion for overwrite
    if driver.lower() == 'geojson':
        try:
            # os.remove(path)
            os.mkdir(path)
            os.chmod(path, int('000', 8))
            with pytest.raises(RuntimeError, match='Could not delete'):
                ds.acreate_vector(path, gtype, ftypes, driver=driver, options=[], sr=SR1['wkt'], ow=True)
        finally:
            os.chmod(path, int('777', 8))
            os.rmdir(path)

def test_mem(driver_mem, gtype, geoms, ftypes, fields):
    driver = driver_mem
    ds = Dataset(allow_none_geometry=1)

    with ds.acreate_vector('', gtype, ftypes, driver=driver, options=[], sr=SR1['wkt']).close as v:
        for geom in geoms:
            v.insert_data(geom, fields)

        # TESTS 0
        assert srs.wkt_same(v.wkt_stored, SR1['wkt'])
        assert srs.wkt_same(v.wkt_virtual, SR1['wkt'])
        for ftype, ftype_stored in zip(ftypes, v.fields):
            for key, value in ftype.items():
                assert ftype_stored[key] == value
        assert v.mode == 'w'
        assert v.type.lower() == gtype.lower()

        # TESTS 1
        datas = list(v.iter_data())
        assert len(datas) == len(geoms)
        assert len(v) == len(geoms)
        # assert v.layer in {0, ''}

        if not sg.GeometryCollection(geoms).is_empty:
            assert eq(
                v.bounds,
                v.bounds_stored,
                v.extent[[0, 2, 1, 3]],
                v.extent_stored[[0, 2, 1, 3]],
                sg.GeometryCollection(geoms).bounds
            )

        for geom, data in zip(geoms, datas):
            if not isinstance(data, tuple):
                data = (data,)
            read_geom, *read_fields = data
            if read_geom is None or read_geom.is_empty:
                assert geom.is_empty
            else:
                assert (geom ^ read_geom).is_empty, (
                    geom.wkt,
                    read_geom.wkt,
                )



from buzzard._a_stored_vector import AStoredVector, ABackStoredVector
from buzzard._a_emissary import AEmissary, ABackEmissary

class AEmissaryVector(AEmissary, AStoredVector):
    """Base abstract class defining the common behavior of all vectors that are backed by a driver.

    Features Defined
    ----------------
    - Has a `layer` (if the driver supports it)
    """

    @property
    def layer(self):
        return self._back.layer

class ABackEmissaryVector(ABackEmissary, ABackStoredVector):
    """Implementation of AEmissaryVector's specifications"""

    def __init__(self, layer, **kwargs):
        super(ABackEmissaryVector, self).__init__(**kwargs)
        self.layer = layer



import collections
import itertools

class MultiOrderedDict(object):
    """Data structure derived from collections.OrderedDict that accept several keys"""

    def __init__(self):
        self._od = collections.OrderedDict()
        self._key_of_ukey = {}
        self._ukeys_of_key = collections.defaultdict(list)
        self._i = 0

    def __str__(self): # pragma: no cover
        l = []
        for key, g in itertools.groupby(map(self._key_of_ukey.get, self._od.keys())):
            l += ['{} x{}'.format(
                key, len(list(g))
            )]
        return '<' + ', '.join(l) + '>'

    def __contains__(self, key):
        return key in self._ukeys_of_key

    def __len__(self):
        return len(self._od)

    def __and__(self, other_set):
        return self._ukeys_of_key.keys() & other_set

    def count(self, key):
        m = len(self._ukeys_of_key[key])
        if m == 0:
            del self._ukeys_of_key[key]
        return m

    def pop_back(self):
        ukey, value = self._od.popitem(last=False)

        key = self._key_of_ukey[ukey]

        ukey_list = self._ukeys_of_key[key]
        assert len(ukey_list) > 0
        assert ukey_list[-1] == ukey
        ukey_list.pop(-1)

        del self._key_of_ukey[ukey]
        if len(ukey_list) == 0:
            del self._ukeys_of_key[key]

        return key, value

    def push_front(self, key, value):
        ukey = self._i
        self._i += 1
        self._ukeys_of_key[key].insert(0, ukey)
        self._key_of_ukey[ukey] = key
        self._od[ukey] = value

    def pop_first_occurrence(self, key):
        if key not in self: # pragma: no cover
            raise KeyError('{} not in MultiOrderedDict'.format(key))

        ukey_list = self._ukeys_of_key[key]
        assert len(ukey_list) > 0
        ukey = ukey_list.pop(0)

        value = self._od[ukey]

        del self._od[ukey]
        del self._key_of_ukey[ukey]
        if len(ukey_list) == 0:
            del self._ukeys_of_key[key]

        return value

    def pop_last_occurrence(self, key):
        if key not in self: # pragma: no cover
            raise KeyError('{} not in MultiOrderedDict'.format(key))

        ukey_list = self._ukeys_of_key[key]
        assert len(ukey_list) > 0
        ukey = ukey_list.pop(-1)

        value = self._od[ukey]

        del self._od[ukey]
        del self._key_of_ukey[ukey]
        if len(ukey_list) == 0:
            del self._ukeys_of_key[key]

        return value

    def pop_all_occurrences(self, key):
        if key not in self:
            return []

        ukey_list = self._ukeys_of_key[key]

        res = [
            self._od[ukey]
            for ukey in reversed(ukey_list)
        ]
        for ukey in ukey_list:
            del self._od[ukey]
            del self._key_of_ukey[ukey]
        del self._ukeys_of_key[key]

        return res



import uuid
import contextlib

from osgeo import gdal

from buzzard._a_pooled_emissary_raster import APooledEmissaryRaster, ABackPooledEmissaryRaster
from buzzard._a_gdal_raster import ABackGDALRaster
from buzzard._tools import conv, GDALErrorCatcher
from buzzard._footprint import Footprint

class GDALFileRaster(APooledEmissaryRaster):
    """Concrete class defining the behavior of a GDAL raster using a file.

    >>> help(Dataset.open_raster)
    >>> help(Dataset.create_raster)

    Features Defined
    ----------------
    None
    """

    def __init__(self, ds, allocator, open_options, mode):
        back = BackGDALFileRaster(
            ds._back, allocator, open_options, mode,
        )
        super(GDALFileRaster, self).__init__(ds=ds, back=back)

class BackGDALFileRaster(ABackPooledEmissaryRaster, ABackGDALRaster):
    """Implementation of GDALFileRaster"""

    def __init__(self, back_ds, allocator, open_options, mode):
        uid = uuid.uuid4()

        with back_ds.acquire_driver_object(uid, allocator) as gdal_ds:
            path = gdal_ds.GetDescription()
            driver = gdal_ds.GetDriver().ShortName
            fp_stored = Footprint(
                gt=gdal_ds.GetGeoTransform(),
                rsize=(gdal_ds.RasterXSize, gdal_ds.RasterYSize),
            )
            channels_schema = self._channels_schema_of_gdal_ds(gdal_ds)
            dtype = conv.dtype_of_gdt_downcast(gdal_ds.GetRasterBand(1).DataType)
            sr = gdal_ds.GetProjection()
            if sr == '':
                wkt_stored = None
            else:
                wkt_stored = sr

        super(BackGDALFileRaster, self).__init__(
            back_ds=back_ds,
            wkt_stored=wkt_stored,
            channels_schema=channels_schema,
            dtype=dtype,
            fp_stored=fp_stored,
            mode=mode,
            driver=driver,
            open_options=open_options,
            path=path,
            uid=uid,
        )

    @contextlib.contextmanager
    def acquire_driver_object(self):
        with self.back_ds.acquire_driver_object(
            self.uid,
            self.allocator
        ) as gdal_ds:
            yield gdal_ds

    def delete(self):
        super(BackGDALFileRaster, self).delete()

        success, payload = GDALErrorCatcher(gdal.GetDriverByName, none_is_error=True)(self.driver)
        if not success: # pragma: no cover
            raise ValueError('Could not find a driver named `{}` (gdal error: `{}`)'.format(
                self.driver, payload[1]
            ))
        dr = payload

        success, payload = GDALErrorCatcher(dr.Delete, nonzero_int_is_error=True)(self.path)
        if not success: # pragma: no cover
            raise RuntimeError('Could not delete `{}` using driver `{}` (gdal error: `{}`)'.format(
                self.path, dr.ShortName, payload[1]
            ))

    def allocator(self):
        return self.open_file(self.path, self.driver, self.open_options, self.mode)

    @staticmethod
    def open_file(path, driver, options, mode):
        """Open a raster dataset"""

        success, payload = GDALErrorCatcher(gdal.OpenEx, none_is_error=True)(
            path,
            conv.of_of_mode(mode) | conv.of_of_str('raster'),
            [driver],
            options,
        )
        if not success:
            raise RuntimeError('Could not open `{}` using driver `{}` (gdal error: `{}`)'.format(
                path, driver, payload[1]
            ))
        gdal_ds = payload

        return gdal_ds



import logging
import functools
import os
import contextlib
import multiprocessing as mp
import multiprocessing.pool

import numpy as np

from buzzard._actors.message import Msg
from buzzard._actors.pool_job import MaxPrioJobWaiting, PoolJobWorking
from buzzard._gdal_file_raster import BackGDALFileRaster
from buzzard._tools import conv
from buzzard._footprint import Footprint

LOGGER = logging.getLogger(__name__)

class ActorFileChecker(object):
    """Actor that takes care of performing various checks on a cache file from a pool"""

    def __init__(self, raster):
        self._raster = raster
        self._back_ds = raster.back_ds
        self._alive = True
        io_pool = raster.io_pool
        if io_pool is not None:
            if isinstance(io_pool, mp.pool.ThreadPool):
                self._same_address_space = True
            elif isinstance(io_pool, mp.pool.Pool):
                self._same_address_space = False
            else: # pragma: no cover
                assert False, 'Type should be checked in facade'
            self._waiting_room_address = '/Pool{}/WaitingRoom'.format(id(io_pool))
            self._working_room_address = '/Pool{}/WorkingRoom'.format(id(io_pool))
        self._waiting_jobs = set()
        self._working_jobs = set()
        self.address = '/Raster{}/FileChecker'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_infer_cache_file_status(self, cache_fp, path):
        msgs = []

        if self._raster.io_pool is None:
            work = Work(self, cache_fp, path)
            status = work.func()
            msgs += [Msg(
                'CacheSupervisor', 'inferred_cache_file_status', cache_fp, path, status
            )]
        else:
            wait = Wait(self, cache_fp, path)
            self._waiting_jobs.add(wait)
            msgs += [Msg(self._waiting_room_address, 'schedule_job', wait)]

        return msgs

    def receive_token_to_working_room(self, job, token):
        self._waiting_jobs.remove(job)
        work = Work(self, job.cache_fp, job.path)
        self._working_jobs.add(work)
        return [
            Msg(self._working_room_address, 'launch_job_with_token', work, token)
        ]

    def receive_job_done(self, job, status):
        self._working_jobs.remove(job)
        return [
            Msg('CacheSupervisor', 'inferred_cache_file_status', job.cache_fp, job.path, status)
        ]

    def receive_die(self):
        assert self._alive
        self._alive = False
        msgs = []
        for job in self._waiting_jobs:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
        for job in self._working_jobs:
            msgs += [Msg(self._working_room_address, 'cancel_job', job)]
        self._waiting_jobs.clear()
        self._working_jobs.clear()
        self._raster = None
        self._back_ds = None
        return msgs

    # ******************************************************************************************* **

class Wait(MaxPrioJobWaiting):
    def __init__(self, actor, cache_fp, path):
        self.cache_fp = cache_fp
        self.path = path
        super().__init__(actor.address)

class Work(PoolJobWorking):
    def __init__(self, actor, cache_fp, path):
        self.cache_fp = cache_fp
        self.path = path
        if actor._raster.io_pool is None or actor._same_address_space:
            func = functools.partial(
                _cache_file_check,
                cache_fp, path, len(actor._raster), actor._raster.dtype,
                actor._back_ds
            )
        else:
            func = functools.partial(
                _cache_file_check,
                cache_fp, path, len(actor._raster), actor._raster.dtype,
                None,
            )
        actor._raster.debug_mngr.event('object_allocated', func)
        super().__init__(actor.address, func)

def _checksum(fname, buffer_size=512 * 1024, dtype='uint64'):
    # https://github.com/earthcube-lab/buzzard/pull/39/#discussion_r239071556
    dtype = np.dtype(dtype)
    dtypesize = dtype.itemsize
    assert buffer_size % dtypesize == 0
    assert np.issubdtype(dtype, np.unsignedinteger)

    acc = dtype.type(0)
    with open(fname, "rb") as f:
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore', r'overflow encountered')

            for chunk in iter(lambda: f.read(buffer_size), b""):
                head = np.frombuffer(chunk, dtype, count=len(chunk) // dtypesize)
                head = np.add.reduce(head, dtype=dtype, initial=acc)
                acc += head

                tailsize = len(chunk) % dtypesize
                if tailsize > 0:
                    # This should only be needed for file's tail
                    tail = chunk[-tailsize:] + b'\0' * (dtypesize - tailsize)
                    tail = np.frombuffer(tail, dtype)
                    acc += tail
        return '{:016x}'.format(acc.item())

def _cache_file_check(cache_fp, path, channel_count, dtype, back_ds_opt):
    checksum = path
    checksum = checksum.split('.')[-2]
    checksum = checksum.split('_')[-1]
    new_checksum = _checksum(path)
    if new_checksum != checksum:
        if back_ds_opt is not None:
            back_ds_opt.deactivate(path)
        LOGGER.warning('Removing {} because invalid checksum ({} instead of {})'.format(
            path, new_checksum, checksum,
        ))
        os.remove(path)
        return False

    allocator = lambda: BackGDALFileRaster.open_file(path, 'GTiff', [], 'r') # This may raise
    with contextlib.ExitStack() as stack:
        try:
            if back_ds_opt is None:
                gdal_ds = allocator()
            else:
                gdal_ds = stack.enter_context(back_ds_opt.acquire_driver_object(path, allocator))

            file_fp = Footprint(
                gt=gdal_ds.GetGeoTransform(),
                rsize=(gdal_ds.RasterXSize, gdal_ds.RasterYSize),
            )
            file_dtype = conv.dtype_of_gdt_downcast(gdal_ds.GetRasterBand(1).DataType)
            file_len = gdal_ds.RasterCount
            if file_fp != cache_fp: # pragma: no cover
                raise RuntimeError('invalid Footprint of {}({} instead of {})'.format(
                    path, file_fp, cache_fp
                ))
            if file_dtype != dtype: # pragma: no cover
                raise RuntimeError('invalid dtype of {}({} instead of {})'.format(
                    path, file_dtype, dtype
                ))
            if file_len != channel_count: # pragma: no cover
                raise RuntimeError('invalid channel_count of {}({} instead of {})'.format(
                    path, file_len, channel_count
                ))
            del gdal_ds
        except Exception:
            # Those exceptions should not trigger a cache file removal, because it might originate
            # from a mistake in the code that does not mean that those files are corrupted. For exemple:
            # - Maximum number of file descriptors reach
            # - Mismatch in cache directories path
            back_ds_opt.deactivate(path)
            raise

    return True



# pylint: disable=redefined-outer-name, unused-argument

import itertools
import operator

import numpy as np
import pytest

from buzzard.test import make_tile_set

PARAMS1 = [
    'extend',
    'overlap',
    'exclude',
    'exception',
    'shrink',
]
PARAMS2 = ['br', 'tr', 'tl', 'bl']
COMBOS = {
    (tx, ty, ox, oy, be)
    for (tx, ty, ox, oy, be) in itertools.product(
        range(4), range(4), range(4), range(4), PARAMS1)
}
FAIL_COMBOS = {
    (tx, ty, ox, oy, be)
    for (tx, ty, ox, oy, be) in COMBOS
    if tx == 0 or ty == 0
    or tx >= 4 or ty >= 4
    or ox >= 3 or oy >= 3
    or (ox == 2 and tx != 1)
    or (oy == 2 and ty != 1)
    or (ox == 1 and tx >= 3)
    or (oy == 1 and ty >= 3)
    or (be == 'exception' and (
        (tx == 2 and ox != 1) or
        (ty == 2 and oy != 1)
    ))
}
VALID_COMBOS = COMBOS - FAIL_COMBOS

# *************************************************************************** **
# FIXTURES ****************************************************************** **
# *************************************************************************** **

@pytest.fixture(scope='module')
def fps():
    """
    See make_tile_set
    A B C D E
    F G H I J
    K L M N O
    P Q R S T
    U V W X Y
    """
    return make_tile_set.make_tile_set(5, [1, -1], [1, -1])


@pytest.fixture(params=PARAMS2)
def boundary_effect_locus(request):
    return request.param


def pytest_generate_tests(metafunc):
    """
    Implementation glossary
    -------
    tx: tile per row
    ty: tile per column
    ox: overlap horizontal
    oy: overlap vertical
    """
    if metafunc.function == test_fail:
        metafunc.parametrize(
            argnames='tx, ty, ox, oy, be',
            argvalues=FAIL_COMBOS,
        )
    if metafunc.function == test_success:
        metafunc.parametrize(
            argnames='tx, ty, ox, oy, be',
            argvalues=VALID_COMBOS,
        )

# *************************************************************************** **
# TESTS  ******************************************************************** **
# *************************************************************************** **

def test_fail(fps, tx, ty, ox, oy, be, boundary_effect_locus):
    with pytest.raises(ValueError):
        fps.GS.tile_count(
            tx, ty, ox, oy,
            boundary_effect=be, boundary_effect_locus=boundary_effect_locus
        )


def test_success(fps, tx, ty, ox, oy, be, boundary_effect_locus):
    tiles = fps.GS.tile_count(
        tx, ty, ox, oy,
        boundary_effect=be, boundary_effect_locus=boundary_effect_locus
    )
    for assert_prop in _BOUNDARY_EFFECT_PROPERTIES[be]:
        assert_prop(fps, tiles, tx, ty, ox, oy, boundary_effect_locus)


def assert_property_tile_size(fps, tiles, tx, ty, ox, oy, param2):
    w = np.vectorize(operator.attrgetter('w'))(tiles.flatten())
    assert np.unique(w).size == 1
    h = np.vectorize(operator.attrgetter('h'))(tiles.flatten())
    assert np.unique(h).size == 1


def assert_property_overlap(fps, tiles, tx, ty, ox, oy, param2):
    if param2 == 'br':
        stride = tiles[0, 0].rsize - (ox, oy)
        refptx, refpty = 'lx', 'ty'
    elif param2 == 'tl':
        stride = tiles[-1, -1].rsize - (ox, oy)
        refptx, refpty = 'rx', 'by'
    elif param2 == 'tr':
        stride = tiles[-1, 0].rsize - (ox, oy)
        refptx, refpty = 'lx', 'by'
    elif param2 == 'bl':
        stride = tiles[0, -1].rsize - (ox, oy)
        refptx, refpty = 'rx', 'ty'
    else:
        assert False
    refptx = operator.attrgetter(refptx)
    refpty = operator.attrgetter(refpty)
    if tiles.shape[0] > 1:
        diffs_vert = np.vectorize(
            lambda a, b: abs(refpty(a) - refpty(b)))(
                tiles[0:-1], tiles[1:])
        assert (diffs_vert == stride[1]).all()
    if tiles.shape[1] > 1:
        diffs_horiz = np.vectorize(
            lambda a, b: abs(refptx(a) - refptx(b)))(
                tiles[:, 0:-1], tiles[:, 1:])
        assert (diffs_horiz == stride[0]).all()


def assert_property_global_bounds(fps, tiles, tx, ty, ox, oy, param2):
    for t in tiles.flatten():
        assert t.poly.within(fps.GS.poly)


def assert_property_share_area(fps, tiles, tx, ty, ox, oy, param2):
    border_tiles = np.r_[tiles[-1, 1:-1], tiles[0, 1:-1], tiles[:, 0], tiles[:, -1]]
    for t in border_tiles:
        assert t.share_area(fps.GS)


def assert_property_full_pixel_coverage(fps, tiles, tx, ty, ox, oy, param2):
    mask = np.zeros(fps.GS.shape, dtype='int')
    for t in tiles.flatten():
        mask[t.slice_in(fps.GS, clip=True)] += 1
    assert (mask > 0).all()

def assert_property_shape(fps, tiles, tx, ty, ox, oy, param2):
    assert tiles.shape == (ty, tx)


def assert_property_unique(fps, tiles, tx, ty, ox, oy, param2):
    tls = np.vectorize(operator.attrgetter('tl'), signature='()->(2)')(tiles.flatten())
    assert np.unique(tls, axis=0).shape[0] == tiles.size


def assert_property_origin(fps, tiles, tx, ty, ox, oy, param2):
    if param2 == 'br':
        assert (tiles[0, 0].tl == fps.GS.tl).all()
    elif param2 == 'tl':
        assert (tiles[-1, -1].br == fps.GS.br).all()
    elif param2 == 'tr':
        assert (tiles[-1, 0].bl == fps.GS.bl).all()
    elif param2 == 'bl':
        assert (tiles[0, -1].tr == fps.GS.tr).all()
    else:
        assert False

_BOUNDARY_EFFECT_PROPERTIES = {
    'extend': [
        assert_property_tile_size,
        assert_property_overlap,
        assert_property_full_pixel_coverage,

        assert_property_shape,
        assert_property_origin,
        assert_property_share_area,
        assert_property_unique,
    ],
    'overlap': [
        assert_property_tile_size,
        assert_property_global_bounds,
        assert_property_full_pixel_coverage,

        assert_property_shape,
        assert_property_origin,
        assert_property_share_area,
        assert_property_unique,
    ],
    'exclude': [
        assert_property_tile_size,
        assert_property_overlap,
        assert_property_global_bounds,

        assert_property_shape,
        assert_property_origin,
        assert_property_share_area,
        assert_property_unique,
    ],
    'shrink': [
        assert_property_overlap,
        assert_property_global_bounds,
        assert_property_full_pixel_coverage,

        assert_property_shape,
        assert_property_origin,
        assert_property_share_area,
        assert_property_unique,
    ],
    'exception': [
        assert_property_tile_size,
        assert_property_overlap,
        assert_property_global_bounds,
        assert_property_full_pixel_coverage,

        assert_property_shape,
        assert_property_origin,
        assert_property_share_area,
        assert_property_unique,
    ],

}

def test_value_error(fps):
    with pytest.raises(ValueError, match='colcount'):
        fps.AI.tile_count(1, -1)
    with pytest.raises(ValueError, match='rowcount'):
        fps.AI.tile_count(-1, 1)
    with pytest.raises(ValueError, match='effect'):
        fps.AI.tile_count(1, 1, boundary_effect='')
    with pytest.raises(ValueError, match='effect_locus'):
        fps.AI.tile_count(1, 1, boundary_effect_locus='')
